""" Pedestals controller

All pedestals management happens here.

Background:

    Pedestals are used to compensate for imperfections in the chip manufacturing
    process, by capturing enough events from the storage array it's possible to
    statistically rule out data originating from these imperfections rather
    than the real world event.

Instructions:

    This controller can run as a one-shot:
    PedestalsController(board, daq).generate_pedestals()

    or persistent:
    ped_ctrl = PedestalsController(board, daq)
    ped_ctrl.generate_pedestals()

.. todo::
    - _capture_data_for_pedestals currently works different for SiREAD, should be uniform.
    - Need a way to track progress and run non-blocking

"""
import copy
import gzip
import logging
import os
import pickle
import time
from collections import deque
from typing import Iterable

import deprecation
import numpy as np

from naludaq.communication import control_registers, digital_registers
from naludaq.controllers import get_board_controller, get_readout_controller
from naludaq.daq import get_daq
from naludaq.helpers.exceptions import (
    NotAValidEvent,
    OperationCanceledError,
    PedestalsDataCaptureError,
)
from naludaq.io import io_manager
from naludaq.tools.metadata import Metadata

LOGGER = logging.getLogger(
    "naludaq.pedestals_controller"
)  # pylint: disable=invalid-name
DigitalRegisters = digital_registers.DigitalRegisters
ControlRegisters = control_registers.ControlRegisters


class PedestalsController:
    """Pedestals controller manages the pedestals generation.

    The pedestals are used to reduce noise originating from the chip itself.
    By capturing data from the board it's possible to use the mean value of x events to
    generate an average error for every sample. Since this average is
    an effect of the imperfections in the chip, subtracting the average counteracts
    the effect of the imperfections.

    The average is improving with 1/sqrt(n).

    This controller relies on an external data acquisiton capturing data to the
    .pedstore buffer it also relies on a board with an active connection to
    communicate with the hardware.

    The pedestals can then be used by the parser to remove the noise.

    Args:
        board:
        num_captures (int): Number of datapoints per sample, used for averaging the values.
        num_warmup_events (int): Number of initial events to discard. Helps the board
            settle and excludes events that may skew the pedestals.
        channels (list[int]): list of channels to generate pedestals for. Any channels excluded
            will result as `np.nan` in the generated pedestals.

    Attributes:
        num_captures: How many datapoints/samplepoint used to calculate the average.

    Raises:
        NotImplementedError if the given board does not support pedestals.
    """

    def __init__(
        self, board, num_captures: int = 10, num_warmup_events: int = 10, channels=None
    ):

        self._board = None
        self.board = board
        # The type of daq used depends on the connection type
        self._daq = get_daq(self.board, parsed=True)

        self.pedestals = dict()
        self.block_buffer = deque()  # Can be used for progress check
        self.validated_data = deque()  # Can be used for progress check
        self._warmup_data = deque()  # each element is a block
        self.validated_blocks = []

        self.channels = channels or list(range(self.board.channels))
        self.num_captures = num_captures
        self.num_warmup_events = num_warmup_events
        self._store_warmup_events = False

        self.margin = 1.1
        self.sleeptime = sleep_calc(self.board, len(self._channels))
        self._progress = []
        self._cancel = False

    @property
    def board(self):
        """Get/set board for pedestals capture.

        Raises:
            NotImplementedError if the given board does not support pedestals.
        """
        return self._board

    @board.setter
    def board(self, board):
        if not board.is_feature_enabled("pedestals"):
            raise NotImplementedError(
                f'Board "{board.model}" does not have support for pedestals.'
            )
        self._board = board

    @property
    def progress(self):
        """Get/Set the progress message queue.

        This is a hook the read the progress if running threads.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError(
                "Progress updates are stored in an object with an 'append' method"
            )
        self._progress = value

    @property
    def channels(self) -> "list[int]":
        """Get/set the channels that will be read out. Any channels not enabled will result
        in NANs for that channel.
        """
        return self._channels

    @channels.setter
    def channels(self, channels: "list[int]"):
        if not isinstance(channels, Iterable):
            raise TypeError("Channels must be a list[int]")
        if len(channels) == 0:
            raise ValueError("At least one channel must be given")
        if any(not isinstance(c, int) for c in channels):
            raise TypeError("Channels must be a list[int]")
        if any(not 0 <= c < self.board.channels for c in channels):
            raise ValueError("One or more channel numbers are out of bounds")
        self._channels = list(set(channels))

        # minimum sleep time changes with number of channels
        self.sleeptime = sleep_calc(self.board, len(self._channels))

    @property
    def metadata(self) -> Metadata:
        """Get a proxy object to use for accessing pedestals metadata."""
        return Metadata(self.board.pedestals)

    def cancel(self):
        """Cancels the pedestals generation as soon as possible.
        No pedestals are generated, and the board is restored to
        its previous state.

        Can only be called from a separate thread.
        """
        self._cancel = True

    def generate_pedestals(self):
        """Generates pedestals and stores them in the board.pedestals.

        Pedestals stored on the board, they will be used on the next acquisition.

        If canceled, the board is set back to its previous state, and no pedestals
        are generated.
        """
        self._cancel = False
        old_peds = getattr(self.board, "pedestals", {})
        self.reset_pedestals()
        self._reset_buffers()
        self._store_board_metadata()
        backup = self._backup_settings()

        self._generate_pedestals_metadata_pre()
        try:
            self._capture_data_for_pedestals()
        except (Exception, KeyboardInterrupt):
            self.cancel()
            raise
        finally:
            self._clean_up(backup)
            if self._cancel:
                self.board.pedestals = old_peds
        if self._cancel:
            return
        self._generate_pedestals_metadata_post()
        self._generate_pedestals_data()
        self._generate_pedestals_from_data()
        _update_progress(self.progress, 100, "Pedestals generated")

    def reset_pedestals(self):
        """Remove the current pedestal data.

        Replaces the pedestals with a set of blank pedestals. All values set
        to zero with the shape set from board params.
        """
        self.board.pedestals = dict()
        self.board.pedestals["data"] = self._set_pedestals_to_zero()
        self._reset_pedestals_data()

    def _set_pedestals_to_zero(self):
        """Set pedestals to zero."""
        output = np.zeros(
            shape=(
                self.board.params["channels"],
                self.board.params["windows"],
                self.board.params["samples"],
            )
        )
        return output

    def _reset_pedestals_data(self, num_captures=None):
        """Reset all pedestals data to an empty shape.

        The shape of the empty pedestals.data is:
        [channels][windows][samples][num_captures]

        Args:
            num_captures (int): Override number of captures for pedestals generation.
                Leave as None is unsure.
        """
        if self.board.pedestals is None:
            self.board.pedestals = dict()
        self.board.pedestals["rawdata"] = self._set_pedestals_raw_to_zero(num_captures)

    def _set_pedestals_raw_to_zero(self, num_captures):
        """Build the structure for raw pedestals data."""
        if num_captures is None:
            num_captures = self.num_captures

        output = np.empty(
            shape=(
                self.board.params["channels"],
                self.board.params["windows"],
                self.board.params["samples"],
                num_captures,
            )
        )
        output[:] = np.NaN

        return output

    def _reset_buffers(self):
        self.block_buffer = deque()
        self.validated_data = deque()
        self.validated_blocks = []
        self._switch_daq_buffer()

    def _clean_up(self, backup):
        """Restores backup settings and stops board and daq readout/capture"""
        self._restore_backup_settings(backup)
        get_board_controller(self.board).stop_readout()
        self._daq.stop_capture()

    def _backup_settings(self) -> dict:
        """Backup settings that might get overwritten to a dict.

        Returns:
            dict with the backup settings:
                'readout_channels',
                'control_registers',
                'digital_registers'
        """
        backup = {
            "readout_channels": get_readout_controller(
                self.board
            ).get_readout_channels(),
            "control_registers": copy.deepcopy(
                self.board.registers.get("control_registers", {})
            ),
            "digital_registers": copy.deepcopy(
                self.board.registers.get("digital_registers", {})
            ),
        }

        return backup

    def _capture_data_for_pedestals(
        self,
        num_captures=None,
        num_warmup_events=None,
    ):
        """Read the data for pedestals generation.

        The data will be captured in blocks, each block will be a fixed set of window numbers.

        Reads `num_captures` samples from each block to average the data.

        The warmup events is always capturing warmup events BEFORE real data.
        The logic will throw away the warmup before transfering validating events.
        in the capture buffer each block will store the data:
         `[warmup_events..., pedestalsdata...]`
        the events in the left will be popped before validation.

        Args:
            num_captures (int): number of events per block
            num_warmup_events (int): number of warmup events to readout. These events
                are not included in the pedestals data.
        """
        board = self.board
        params = board.params

        LOGGER.debug(
            "Board is %s with block: %s", params["model"], params["pedestals_blocks"]
        )
        total_windows = params["windows"]
        block_size = params["pedestals_blocks"]

        # Validation in case of pedestals_blocks typo
        block_size = min(max(block_size, 1), total_windows)

        # Reset readoutchannels to all
        rc = get_readout_controller(board)
        try:
            rc.set_readout_channels(self._channels)
            rc.set_read_window(write_after_trig=total_windows)
            rc.set_read_window(windows=block_size)
        except Exception as e:
            LOGGER.debug("Setting the windows failed due to %s", e)

        num_blocks = self._calculate_number_blocks(block_size, total_windows)

        # Extra code for status #############################################
        min_value = 0
        max_value = 80
        step_value = (max_value - min_value) / num_blocks
        #####################################################################
        LOGGER.debug("Capturing data for pedestals")

        # This will break the child threads and stop capture in cmd-line and notebooks.
        for block in range(num_blocks):
            try:
                prog = min(min_value + step_value * block, max_value)
                _update_progress(
                    self.progress, prog, f"Capturing block: {block+1}/{num_blocks}"
                )

                LOGGER.debug("Block %s/%s", block, num_blocks)

                # Capture validated data for one block into self.validated_data
                if not self._capture_block(
                    block, block_size, num_captures, num_warmup_events
                ):
                    raise PedestalsDataCaptureError(
                        "Failed to validate the return data"
                    )

                # Transfer into validated_blocks, used to generate peds raw data
                self.validated_blocks.append(self.validated_data)

                if self._cancel:
                    break
            except KeyboardInterrupt:
                LOGGER.info("Caught keyboard interrupt, stopping capture children")
                raise KeyboardInterrupt

    def _calculate_number_blocks(self, block_size: int, windows: int) -> int:
        """Calculates number of blocks to use with a certain block_size.

        If blocksize doesn't divide evenly with number of windows, then add one more
        block to capture data the rest.

        Args:
            block_size(int): Amount of windows in a block.
            windows(int): Amount of windows of the hardware.

        Returns:
            Number of blocks as an integer.
        """
        num_blocks = windows // block_size + (windows % block_size > 0)
        return num_blocks

    def _switch_daq_buffer(self):
        """Set the daqs output buffer to the pedestals controllers storage.

        Tell the daq pipeline to store the events for the pedestals in
        a separate storage for the pedestals.
        """
        self._daq.output_buffer = self.block_buffer

    def _capture_block(self, block, block_size, num_captures, num_warmup_events=None):
        """Captures a single block, only keeping validated events.

        Args:
            block (int): the block index (0 to windows / block count - 1)
            block_size (int): number of windows per block
            num_captures (int): number of captures that contribute to pedestals data.
            num_warmup_events (int): number of events to readout before capturing pedestals
                data. Warmup events are not used to generate pedestals.

        Raises:
            TimeoutError if the data could not be read back.

        Returns:
            True if successful, or False otherwise.
        """
        # Readout settings
        num_captures = num_captures or self.num_captures
        num_warmup_events = num_warmup_events or self.num_warmup_events
        total_captures = num_captures + num_warmup_events
        readout_settings = {
            "trig": "i",  # Starts an immediate readout.
            "lb": "f",
            "acq": "ped",
            "ped": "zero",
            "readoutEn": True,
            "singleEv": True,
        }

        self.block_buffer = deque(maxlen=total_captures)
        self.validated_data = deque(maxlen=num_captures)
        self._daq.output_buffer = self.block_buffer
        bc = get_board_controller(self.board)
        rc = get_readout_controller(self.board)
        rc.set_read_window(lookback=block * block_size)

        reset_count = 0
        t_count = 0
        timeout = 10
        needed_amount = total_captures
        while len(self.validated_data) < num_captures:
            if needed_amount <= 0 or self._cancel:
                break

            rc.number_events_to_read(needed_amount)

            self._daq.start_capture()
            bc.start_readout(**readout_settings)
            try:
                finished = self._wait_for_data(self.block_buffer, needed_amount)

                if not finished:
                    if reset_count >= timeout:
                        raise TimeoutError(
                            "Something wrong with setup/hardware, reset 5 times."
                        )
                    LOGGER.debug("data capture failed, restarting")
                    reset_count += 1
                    self.margin = increase_margin(self.margin)
                    self.sleeptime = sleep_calc(self.board, self.margin)

                LOGGER.debug("Data capture finished, stopping")
            finally:
                self._daq.stop_capture()
                bc.stop_readout()
            # self._daq.stop_workers()
            LOGGER.debug("Only keep validated events, need: %s", needed_amount)
            expected_block = self._calculate_expected_block(block, block_size)
            all_validated = False
            if (
                len(self.block_buffer) > num_warmup_events
            ):  # If the amount captured is less than warmup, save none
                all_validated = self._only_keep_validated_events(
                    self.block_buffer,
                    self.validated_data,
                    self._warmup_data if self._store_warmup_events else None,
                    expected_block,
                    num_captures,
                    num_warmup_events,
                )

            if all_validated is True:
                return True
            if t_count > timeout:
                return False
            t_count += 1

            needed_amount = (
                num_warmup_events
                + (num_captures - len(self.validated_data))
                - len(self.block_buffer)
            )

        return True

    def _wait_for_data(self, output_buffer, desired_amount):
        """Wait for the desired amount of data to arrive.

        Returns in these cases:
        - if desired amout is reached
        - if no data is received
        - if too much data is received

        Args:
            output_buffer (deque): storage for data we wait for.
            desired_amount(int): The amount of events we wait for.

        Returns:
            True if the correct amount of data is captured, else False
        """
        result = True
        prev_amount = -1
        # wait for readout to finish
        while len(output_buffer) != prev_amount:
            stime = self.sleeptime * (desired_amount - len(output_buffer))
            if output_buffer:
                prev_amount = len(output_buffer)
            try:  # If output_buffer is larger than desired, stime < 0.
                time.sleep(stime)
            except ValueError:
                pass
            if len(output_buffer) > desired_amount:
                break

        if len(output_buffer) == 0:
            result = False

        return result

    def _only_keep_validated_events(
        self,
        to_validate: deque,
        validated: deque,
        warmup_output: deque,
        expected_block: list,
        expected_amount: int,
        warmup_amount: int,
    ) -> bool:
        """Loop through captured events and only keep validated.

        Args:
            to_validate(deque): events waiting to be validated
            validated(deque): events validated
            expected_block(list): a list of expected window numbers
            expected_amount(int): the amount of events needed to pass.
            warmup_events(int): amount of warmup events before real data.

        Returns:
            True if the expected amount is validated. False if timeout.
        """
        LOGGER.debug("Only keeping validated events.")
        for _ in range(warmup_amount):
            try:
                _ = to_validate.popleft()
            except IndexError:
                return False

        while to_validate and len(validated) != expected_amount:
            try:
                event = to_validate.pop()  # pop from right discards warmup events
            except IndexError:
                continue

            try:
                if self._validate_event(event, expected_block):
                    validated.append(event)
                else:
                    LOGGER.debug("Event not validated")
            except NotAValidEvent as error_msg:
                LOGGER.warning(error_msg)
                LOGGER.debug("Event not validated")

        if len(validated) == expected_amount:
            if warmup_output is not None:
                warmup_output.append(to_validate)
            return True
        return False

    def _validate_event(self, event, expected_block):
        """Returns true if the window labels matches the expected_block.

        The validation matches the received window_labels with the expected
        window labels to make sure the block contains only data from that
        block. The firmware buffer sometimes contains more data from a previous block
        this makes sure it doesn't enter the next block buffer.

        Args:
            event(dict): event to validate
            expected_block(list): a list of expected window numbers

        Returns:
            True if validated, False if the events windows doesn't match expected.

        Raises:
            NotAValidEvent if the event doesn't have a window_labels key.
        """
        test_channel = self._channels[0]
        try:
            if np.all(event["window_labels"][test_channel] == expected_block):
                return True
        except Exception as error_msg:
            raise NotAValidEvent(f"Event validataion failed due to: {error_msg}")
        else:
            # LOGGER.debug(
            print(
                "Expected: %s != returned: %s",
                expected_block,
                event["window_labels"][test_channel],
            )
            return False

    def _calculate_expected_block(self, block, block_size):
        """If a board doesn't have a junk register, there will be more expected blocks.

        Args:
            block(int): Current block number.
            block_size(int): Number of windows captured per block.

        Returns:
            List of expected window numbers.
        """
        expected_block = np.arange(block * block_size, (block + 1) * block_size) % (
            self.board.params["windows"]
        )

        return expected_block

    def _restore_backup_settings(self, backup_settings):
        """Restore all backuped settings to the board.

        Returns:
            True if settings have been restored, False if no old settings were found.
        """
        if not backup_settings:
            return

        get_readout_controller(self.board).set_readout_channels(
            backup_settings["readout_channels"]
        )

        # Restore backup
        for register in ["control_registers", "digital_registers"]:
            self.board.registers[register] = backup_settings.get(register, {})
        DigitalRegisters(self.board).write_all()
        ControlRegisters(self.board).write_all()

    # Metadata #######################################################################################
    def _generate_pedestals_metadata_pre(self):
        """Adds some metadata to the pedestals dict. Called immediately
        before the pedestals data will be captured.
        """
        self._store_board_metadata()
        self._store_sensor_readings()

    def _generate_pedestals_metadata_post(self):
        """Adds some metadata to the pedestals dict. Called immediately
        after the pedestals data has been captured.
        """
        self._store_sensor_readings()
        self._store_time_metadata()

    def _store_board_metadata(self):
        """Store board params/registers into pedestals metadata"""
        self.metadata.set_configuration(self.board)

    def _store_sensor_readings(self):
        """Store sensor readings into pedestals metadata"""
        LOGGER.debug("Storing sensor metadata")
        self.metadata.store_sensor_readings(self.board)

    def _store_time_metadata(self):
        """Store event times into pedestals metadata"""
        LOGGER.debug("Storing time metadata")
        md = self.metadata
        for block in self.validated_blocks:
            md.store_event_times(block)

    # Process data ###################################################################################
    def _generate_pedestals_from_data(self):
        """Generate pedestals data and average it.

        Takes the generated pedestals data and generates the actual pedestals.
        The pedestals are stored in the controller and can be retreived with
        get_pedestals().

        """
        rawdata = self.board.pedestals["rawdata"]
        output = np.nanmean(rawdata, axis=3)
        self.board.pedestals["data"] = output
        if self._store_warmup_events:
            LOGGER.debug("Warmup storage flag set, storing in pedestals")
            self.board.pedestals["warmup_data"] = self._warmup_data

    def _generate_pedestals_data(self, num_captures=None):
        """Generates the pedestals data from captured data.

        Uses captured events in the pedestals_raw_data deque
        to generate the data.

        Args:
            nCaptures (int): Number of captures, default 10
            blocksize (int): size of blocks, default 16
        """

        if num_captures is None:
            num_captures = self.num_captures

        block_size = self.board.params["pedestals_blocks"]

        self._reset_pedestals_data(num_captures)

        LOGGER.debug(
            "Generating pedestals from %s samples with blocksize: %s",
            num_captures,
            block_size,
        )

        return self._create_raw_pedestals_data(block_size, num_captures)

    def _create_raw_pedestals_data(self, block_size, num_captures):
        """Move data to board.pedestals after validating.

        Since data already is prevalidated this is just a logic check,

        Args:
            block_size(int): windows per block
            num_captures(int): amount of events averaged.
        """
        self.board
        num_blocks = self._calculate_number_blocks(
            block_size, self.board.params["windows"]
        )
        assert num_blocks == len(self.validated_blocks)

        # STATUS UPDATES ####################################
        min_value = 80
        max_value = 95
        step_value = (max_value - min_value) / num_blocks
        #####################################################
        for block_idx, block in enumerate(self.validated_blocks):
            LOGGER.info("Capturing block %s/%s", block_idx + 1, num_blocks)
            # Status message ###############################
            _update_progress(
                self.progress,
                min_value + step_value * block_idx,
                f"Processing data: {block_idx+1}/{num_blocks}",
            )
            ###############################################
            for cap, event in enumerate(block):
                for chan in range(self.board.params["channels"]):
                    if chan not in self._channels:
                        continue
                    for window_num in range(block_size):
                        window = window_num + block_idx * block_size

                        # Avoid rolling over and overwrite block 0
                        if block_idx != 0 and (window < block_size):
                            continue
                        # avoid window number rolling over
                        if window >= self.board.params["windows"]:
                            continue

                        for sample in range(self.board.params["samples"]):
                            index = sample + window_num * self.board.params["samples"]
                            data = event["data"][chan][index]
                            self.board.pedestals["rawdata"][chan][window][sample][
                                cap
                            ] = data

        return True

    def _raise_if_canceled(self):
        """Raise an ``OperationCanceledError`` if the cancel flag is set."""
        if self._cancel:
            raise OperationCanceledError("Pedestals generation was canceled.")

    def _increase_sleeptime(self, needed_amount):
        """Increase the sleeptime based on received vs expected.

        Formula is intended to change the margin incrementally without
        causing to much oscillation.
        """
        # inc = (needed_amount / len(self.block_buffer))/2 + 1 settles at 0.45
        # inc = (needed_amount / len(self.block_buffer))/2 fails
        # inc = (needed_amount / len(self.block_buffer)) Oscillates
        block_len = max(len(self.block_buffer), 1)
        inc = (needed_amount / block_len) / 3 + 1  # Settles at 0.38
        self.margin = increase_margin(self.margin, margin_inc=inc, max_margin=10)
        self.sleeptime = sleep_calc(
            self.board, self.margin, channels=len(self._channels)
        )
        LOGGER.debug("Waittime changed to %s", self.sleeptime)

    # Functions to export data
    @staticmethod
    def save_pedestals(pedestals, filename):  # TODO(v.0.1.23): Use IO module.
        """Save the pedestal in binary format for backwards compatibility."""
        if not isinstance(pedestals, dict):
            raise TypeError(f"pedestals must be a dict, got {type(pedestals)}")
        if filename is None:
            raise TypeError("Supplied pathname is NoneType.")
        path, _ = os.path.split(filename)
        if not os.path.isdir(path):
            raise NotADirectoryError(f"Not a valid directory: {path}")

        try:
            sfile = gzip.GzipFile(filename, "w", compresslevel=4)
            pickle.dump(pedestals, sfile, protocol=pickle.HIGHEST_PROTOCOL)
        except IOError as error_msg:
            LOGGER.error("File could not be created. %s", error_msg)
        except pickle.PicklingError as error_msg:
            LOGGER.error("Saving pedestals failed: %s", error_msg)

    def load_pedestals(self, filename):  # TODO(v.0.1.23): Use IO module.
        """Load the pedestal gziped and pickled.

        The pedestals object is loaded to both self.pedestals
        and is returned.

        Args:
            filename: valid filename

        Returns:
            Pedestals data.
        """
        outped = get_pedestals_from_file(filename)
        self.board.pedestals = outped
        return outped

    @deprecation.deprecated(
        deprecated_in="0.1.22",
        details="Use the IO Controller for all csv imports/exports",
    )
    def export_pedestals_csv(self, pedestals, filename):
        """Export all data from the pedestals to CSV.

        Please note this will only export the data, not the information regarding the pedestals.
        This function also assumes filename is in the correct format.
        It will override any existing file.

        Args:
            pedestals (Pedestals): A pedestals object with the pedestals data.
            filename (path): Filename fro the csv file.
        """

        io_manager.IOManager().export_pedestals_csv(self.board, pedestals, filename)


def increase_margin(margin: float, margin_inc: float = 1.5, max_margin: float = 5):
    """Increases the margin, using an exponential standoff"""
    margin *= margin_inc
    margin = min(margin, max_margin)
    return margin


def sleep_calc(board, margin: float = 2.0, channels=None) -> float:
    """Compute the `sleeptime` for the board.

    Using the `board.params` to gather blocksize, baudrate and readout
    parameters compute the minumum sleep time.

    The `margin` can be dynamically altered as a standoff in case the calc
    doesn't match real transfer rate.
    By increasing `margin` it's possible to increase `sleeptime` in case of
    real-world differences between calculated baudrate and real. Or if
    the return data is longer/padded.

    Args:
        board(naludaq.board)
        margin(float): Multiplier to increase sleeptime due to real-world
            scenarios.
        channels (int): number of channels to be read on the board.

    Returns:
        sleeptime in seconds as a float

    """
    if channels is None:
        channels = board.channels
    baudrate = board.connection_info.get("speed", 115200)
    transferrate = baudrate // 8
    windows = board.params.get("pedestals_blocks", 16)
    samples = board.params.get("samples", 64)
    data = windows * channels * samples * margin * 2  # 16-bits per value
    sleeptime = data / transferrate

    return sleeptime


def get_pedestals_from_file(filename):
    """Load the pedestal gzipped and pickled.

    The pedestals object is returned.

    Args:
        filename: valid filename

    Returns:
        Pedestals data.
    """
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"No file found: {filename}")

    try:
        sfile = gzip.GzipFile(filename, "r")
        outped = pickle.load(sfile)
    except EOFError as error_msg:
        LOGGER.error("Reached the end of the file, msg: %s", error_msg)
    except IOError as error_msg:
        LOGGER.error("File could not be loaded. %s", error_msg)
    except pickle.UnpicklingError as error_msg:
        LOGGER.error("Loading pedestals failed: %s", error_msg)
    else:
        if not isinstance(outped, dict):
            raise TypeError(f"Not a valid Pedestals file: {filename}")

        return outped


def _update_progress(receiver, percent: float, message: str):
    """Updates a progress "receiver" with a percent and message.

    The receiver can be a list or deque of (percent, message) tuples,
    or a ProgressDialog (for use with NaluScope).

    Args:
        receiver (list, deque, or ProgressDialog): the receiver of the progress update.
            Can be None to report nothing
        percent (float): the percent completion of the task
        message (str): a description of what is currently taking place
    """
    if receiver is None:
        LOGGER.deubg("%s | %s", percent, message)
        return
    elif isinstance(receiver, (list, deque)):
        receiver.append((percent, message))
    else:
        try:
            receiver.update_status(percent, message)
        except:
            pass



"""
There are some other fun things on the boards/ASICs that are useful to control

This controls those things

Ben Rotter
"""
import logging
import time

from naludaq.communication import readI2cResponse, sendI2cCommand
from naludaq.controllers.controller import Controller

logger = logging.getLogger("naludaq.peripherals")


class PeripheralsController(Controller):
    """The peripherals controller handles all the telemetry devices on the boards.

    Telemetry found on the boards is temperature, current, voltages, it also measures certain
    voltage rails.

    The specific telemetry available is highly board dependent.
    """

    def __init__(self, board):
        super().__init__(board)
        self._reader_functions = {
            "temp": self.read_temp_of_chip,
            "current": self.read_current_of_chip,
            "vadjn": self.read_vadjn,
            "vadjp": self.read_vadjp,
        }

    def read_temp_of_chip(self):
        """Reads out the AT30TS00 temp sensor and converts it to Celcius.

        Requires configuration, done in board_controller.init_board()

        Returns:
            Temperature of the chip in Celcius.
        """
        temp = 0
        response = self._request_temperature_readout()
        if response:
            temp = self._convert_response_to_temperature(response)

        logger.debug("Temperature: %s C", temp)

        return temp

    def _request_temperature_readout(self):
        """Reads the temperature register on the AT30TS00.

        Returns:
            A list of binary strings (including ACK bit).
        """
        board = self.board
        sendI2cCommand(self.board, "30", ["05"])
        sendI2cCommand(board, "31", ["FF", "FF"])
        return readI2cResponse(board)

    def _convert_response_to_temperature(self, response) -> float:
        """Converts the AT30TS00 temperature register response
        into a temperature in Celsius

        Args:
            response (list): I2C response registers.

        Returns:
            Temperature in Celsius.
        """
        reg_bits = response[1][:-1] + response[2][:-1]
        temp_bits = reg_bits[3:-1]
        sign_bit = temp_bits[0]

        # Negative temperatures are represented in 2's complement
        temp = int(temp_bits, 2)
        if sign_bit == "1":
            temp -= 1 << len(temp_bits)

        return temp / 8

    def read_current_of_chip(self):
        """Reads out the MCP3426 monitoring the current and converts to Ampere.

        Reads out the MCP3426 monitoring the current sense resistor, and converts
        it to a current in A.

        Returns:
            Current in Ampere.

        Raises:
            NotImplementedError if this functionality is not available for the current model.
        """
        voltage = self._request_readout("current")
        current = self._convert_voltage_to_current(voltage)
        logger.debug("Current Sense: %sA / %smV", current, voltage)
        return current

    def read_vadjn(self):
        """Read the vadjn voltage

        Raises:
            NotImplementedError if this functionality is not available for the current model.
        """
        voltage = self._request_readout("vadjn")
        logger.debug("VadjN: %smV", voltage)
        return voltage

    def read_vadjp(self):
        """Read the vadjp voltage

        Raises:
            NotImplementedError if this functionality is not available for the current model.
        """
        voltage = self._request_readout("vadjp")
        logger.debug("VadjP: %smV", voltage)
        return voltage

    def _request_readout(self, param):
        params = self._get_peripheral_params(param)
        return self._read_MCP3426_ADC(**params)

    def _get_peripheral_params(self, name: str) -> dict:
        """Gets the peripherals parameters for a given type of measurement.

        Args:
            param (str): the measurement name
        """
        try:
            params = self._board.params["peripherals"][name]
        except KeyError:
            raise NotImplementedError(
                f"{name} readout functionality has not been implemented for {self.board.model}"
            )
        return params

    def _read_MCP3426_ADC(self, chan, addr, bits=16, gain=1):
        """
        MCP3426/7/8, two on board (addr D0 and D8)

        command bits: RCCOSSGG
        R = "Ready Bit," initiates new conversion in one-shot mode
                         resets to 1 when new conversion is settled
        CC = Channel Selection
        O = Conversion Mode (1=continuous, 0=one-shot)
        SS = "Sample Rate" - b'00' 240sps (12bit)
                             b'01' 60sps  (14bit)
                             b'10' 15sps  (16bit)
        GG = "Gain Select" - b'00' 1x, b'01' 2x,
                             b'10' 4x, b'11' 8x
        """
        board = self.board

        bitmap = {
            16: "10",
            14: "01",
            12: "00",
        }.get(bits, None)
        if bitmap is None:
            raise ValueError("bits needs to be 16, 14, or 12, not %s", bits)

        gainmap = {
            1: "00",
            2: "01",
            4: "10",
            8: "11",
        }.get(gain, None)
        if gainmap is None:
            raise ValueError("gain needs to be 1, 2, 4, or 8, not %s", gain)

        command = "1" + "{0:b}".format(chan).zfill(2) + "1" + bitmap + gainmap
        sendI2cCommand(board, addr, [command])
        time.sleep(0.1)
        readaddr = "{0:X}".format(addr + 1)
        sendI2cCommand(board, readaddr, ["FF", "FF"])
        time.sleep(0.1)
        response = readI2cResponse(board)

        logger.info("response=%s", response)

        return self._convert_response_to_voltage(response, bits=bits, gain=gain)

    @staticmethod
    def _convert_response_to_voltage(response, bits, gain):
        """When the MCP3426 is operating in 16-bit mode, convert response to voltage.

        When in 16-bit mode LSB represents 62.5 uV. It is also using 8x gain.
        """
        step_size = 4096 / (2**bits)
        return (step_size / gain) * (
            int(response[1][:-1], 2) * (2**8) + int(response[2][:-1], 2)
        )

    def _convert_voltage_to_current(self, voltage):
        """Convert voltage to current by checking voltage drop over resistor.

        # Currently using a 50 mili-ohm resistor.
        # voltage is in mV, this should yield mA
        """

        resistance = self._get_v2a_resistance()
        return voltage / resistance

    def _get_v2a_resistance(self):
        """Default used to be 50mOhm"""
        return self.board.params.get("peripherals", {}).get("current_resistor", 50e-3)

    def read_all(self) -> dict:
        """Read all peripherals into a dict"""
        result = {}
        for name, func in self._reader_functions.items():
            try:
                value = func()
            except Exception:
                value = None
            result[name] = value
        return result



"""A collection of Errors used by the Naludaq package.

An error in the `naludaq` package should derive from the NaludaqException which is the base
exception. By Having a Common base class it's easier to capture and trace exceptions
generated by the package.

"""


class NaludaqException(Exception):
    """Base class for Naludaq exceptions, for tracability."""


class InvalidParameterFile(NaludaqException):
    """The supplied parameter file is not valid."""


class InvalidFileType(NaludaqException):
    """Rased if the filetype is not recognized."""


class BoardConnectionError(NaludaqException):
    """Raised when the board connection unexpectedly failed."""


class InvalidBoardModelError(NaludaqException):
    """Raised when the boardmodel doesn't exist in the list of models."""


class BadDataError(NaludaqException):
    """Captured data is not a valid event."""


class NotAValidEvent(NaludaqException):
    """Event couldn't be validated.

    This Event is not what the program excepted.
    """


class NoDefaultRegistersError(NaludaqException):
    """This board model have no default registers. User must load a file."""


class ClockFileError(NaludaqException):
    """Board needs a clock file but None is loaded."""


class InvalidClockFileError(NaludaqException):
    """Clock file contains invalid clock data"""


class RegisterError(NaludaqException):
    """Generic register Error"""


class RegisterNameError(RegisterError):
    """Not a valid register name."""


class RegisterFileError(RegisterError):
    """Board needs a valid parameter file."""


class InvalidRegisterError(RegisterError):
    """Not a valid register"""


class IncompatibleParserError(NaludaqException):
    """Parser is not compatible with the selected protocol/board model."""


class PedestalsDataCaptureError(NaludaqException):
    """Couldn't capture pedestals data correctly.

    Either due to bad returning blocks or no data captured at all.
    """


class BoardParameterError(NaludaqException):
    """The board paramters lack enough fields to describe the board.

    This could be due to a faulty or old parameter/register file.
    """


class DaqStateError(NaludaqException):
    """The DAQ is in an invalid state for some operation.

    This could be due to workers already being initialized, started,
    or stopped.
    """


class InvalidTriggerMaskError(NaludaqException):
    """The trigger mask supplied is invalid."""


class NoLinearRegionFoundError(NaludaqException):
    """No linear region was found from DAC sweep data.

    This could be due to a poor dac sweep, or too strict parameters
    set in determining the linear region.
    """


class NoCalibrationDataFoundError(NaludaqException):
    """No calibration data attribute found on the board.

    This could be due to the calibration has not been run or the calibration data is corrupted.
    """


class FTDIError(NaludaqException):
    """A problem occurred while interacting with the FTDI drivers or a device
    connected through FTDI.
    """


class DACError(NaludaqException):
    """There was a problem with setting the DACs."""


class I2CError(NaludaqException):
    """Failed to read or write an I2C device."""


class I2CError(NaludaqException):
    """Communication with device through I2C failed."""


class AutoActionError(NaludaqException):
    """The AutoAction tool encountered an error"""


class AutoTriggerError(AutoActionError):
    """TheAutoTrigger tool encountered an error"""


class OperationCanceledError(NaludaqException):
    """The operation was canceled."""


class VersionError(NaludaqException):
    """Invalid or unsupported version"""



"""
Controls how the analog portion of the UDC chip is biased. 
The chip can either use a current mirror to bias the amplifiers (current mode)
or an internal DAC to set the bias voltage (voltage mode).

You can also choose which side to set, so the left can be one mode, and the right could be the other,
if you need to.

Current mode is presumed to be more stable, however the bias value cannot be changed
Voltage mode is presumed to be less stable, but allows for the flexibility of changing the bias voltage

NOT to be confused with External DAC biasing. This biasing affects the stability and performance of the
amplifiers IN the chip. Has nothing to do with external DAC biasing for the input signal.

Kenneth Lauritzen
"""

from naludaq.helpers.exceptions import InvalidBoardModelError


def get_biasing_mode_controller(board):
    """Retrieves the appropriate biasing mode controller.
    Currently, only the UDC16 supports different biasing modes.

    Args:
        board (Board): board to change biasing mode
    """

    if board.model in ["udc16"]:
        from .udc16 import BiasingModeControllerUDC16

        return BiasingModeControllerUDC16(board)

    raise InvalidBoardModelError("Only UDC16 supports biasing mode")



"""Analog registers
===================
Registers controlling the analog part of the ASIC.

What are the analog registers?
It's a set of registers on the FPGA matching a set of registers on the chip
By changing these registers it controls the analog portion of the chip.

Command structure:
------------------
Updating the analog command uses a format 'BXMAAVVV'

- B tells the FPGA command parser to expect a analog/digital command.
- M is mode, 0 = write, 1 = read
- AA = address
- VVV = value

Will be interpereted by the FPGA as hex. it's 32-bit = 8 hex values.

Functions:
----------
- write_all
- write
- write2addr


"""
from logging import getLogger

from naludaq.communication import registers

LOGGER = getLogger(__name__)


class AnalogRegisters(registers.WriteRegisters):
    def __init__(self, board):
        super().__init__(board, "analog_registers")

        self._write_cmd = (
            0xB00  # Standard analog write address, changes with board/mode.
        )
        self.width_cmd = 3
        self.width_addr = 2  # hex char
        self.width_val = 3  # hex char

    # READ #########################################################

    def _get_value_from_addr(self, addr):
        """Generate the value from the hex address.

        An address can contain multiple registers.
        Generate the value depending on the bit positions of all registers on the address.

        Args:
            addr(str): Hex value of the address, 2 hex char long.

        Returns:
            integer value of the combined registers.

        Raises:
            AttributeError if the address is not valid.
        """
        addr = addr.upper()
        value = 0
        failed = True
        strobe_correction_keys = self.board.params.get("strobe_correction_keys", [])
        strobe_values_correction = self.board.params.get(
            "strobe_values_correction", False
        )
        for key, val in self.registers.items():

            if val["address"].upper() != addr:
                continue
            failed = False
            key_value = val["value"]
            key_pos = val["bitposition"]
            key_width = val["bitwidth"]

            # FIX TIMING REGS, ACCORDING TO BOARD PARAMETER timing_correction
            if strobe_values_correction and key in strobe_correction_keys:
                key_value = self._fix_timing(key_value)

            value += 2**key_pos * (int(key_value) % (2**key_width))
        if failed is True:
            raise AttributeError("Address is not found")

        return value

    def _generate_write_request(self, addr: str, value: int) -> str:
        """Generate the string to send to the board.

        OVERLOAD the standard way of generating the command since the analog command parsing
        sometimes

        Abstracts the model differences away from the main code, AARDVARCv2 got a timing hack.

        Args:
            board(obj)
            name(str): name of the analog register
            value(): Value to set the register to.
        """

        write_addr = f"{self._write_cmd:X}"
        cmd_addr = addr

        if len(addr) == 3:
            write_addr = f"B{addr[0]}0"
            cmd_addr = f"{addr[1:]}"

        if isinstance(value, (int, bool)):
            value = f"{value:X}".zfill(self.width_val)

        command = f"{write_addr}{cmd_addr}{value}"  # First 5 hex chars, address should be 3 hex. B001c

        return command

    @staticmethod
    def _fix_timing(in_value: int):
        """The timing signal data mux is all mixed up for simplicity of routing, so this fixes it

        Don't use unless you know what it does. Contains hardcoded values.
        """
        if in_value < 64:
            return in_value * 2 + 1
        return 254 - 2 * in_value



"""Pedestals generator for UPAC96
"""
import logging
import time
from collections import deque

import numpy as np

from naludaq.controllers import get_board_controller
from naludaq.controllers.board import get_board_controller
from naludaq.helpers.exceptions import (
    OperationCanceledError,
    PedestalsDataCaptureError,
    RegisterNameError,
)
from naludaq.tools.pedestals.pedestals_controller import (
    PedestalsController,
    _update_progress,
    sleep_calc,
)
from naludaq.tools.waiter import EventWaiter

LOGGER = logging.getLogger("naludaq.upac96_pedestals_generator")


def disable_trigger_monitor_signal(func):
    """Decorator to disable the trigger monitor signal during the execution of a function.

    Args:
        func (function): function to decorate

    Returns:
        function: decorated function
    """

    def wrapper(self, *args, **kwargs):
        try:
            previous = self.board.registers["control_registers"][
                "trigger_monitor_disable"
            ]["value"]
        except KeyError:
            raise RegisterNameError("trigger_monitor_disable")
        self._board_controller.set_trigger_monitoring_disabled(disabled=True)
        result = func(self, *args, **kwargs)
        self._board_controller.set_trigger_monitoring_disabled(previous)
        return result

    return wrapper


class Upac96PedestalsGenerator(PedestalsController):
    """Pedestals generator for UPAC96.

    There are a couple differences in this pedestals generator:
    - There is no computed sleep time for data capture. Instead, an event waiter is used to stop
      as soon as an event is read.
    - The first window of all channels for each event is thrown away, leaving behind a randomly-placed
      hole in the pedestals data. The pedestals generator continues to capture events beyond the
      specified number of events (``num_captures``) until all necessary data is gathered.
    """

    def __init__(
        self,
        board,
        num_captures: int = 10,
        num_warmup_events: int = 10,
        channels: list[int] = None,
    ):
        super().__init__(
            board,
            num_captures,
            num_warmup_events=num_warmup_events,
            channels=channels,
        )
        self._board_controller = get_board_controller(board)
        self._timeout = sleep_calc(board, margin=2, channels=len(self._channels)) + 3
        self._attempts = 5

    # Capture methods =========================================================================
    @disable_trigger_monitor_signal
    def _capture_data_for_pedestals(
        self,
        num_captures: int = None,
        num_warmup_events: int = None,
    ):
        """Capture data needed for pedestals and store it in the output buffer.

        For UPAC96 it's a bit different. We need to throw away the first window and there's no forced mode,
        so it's random where the window hole shows up in the data. This capture function makes sure that
        regardless of where the hole shows up in each event, we have enough captures for all windows.

        Args:
            num_captures (int): number of data points per sample that contribute to the pedestals data.
            num_warmup_events (int): number of warmup events to capture that do not contribute to pedestals data.
        """
        channels = self._channels
        num_captures = num_captures or self.num_captures
        num_warmup_events = num_warmup_events or self.num_warmup_events
        needed_amount = num_warmup_events + num_captures
        current_amount = 0
        valid_events = []

        LOGGER.debug("Running pedestals generation...")
        self._start_capture()
        daq_buffer = self._daq.output_buffer
        while current_amount < needed_amount:
            LOGGER.info(
                "Capturing next event. Have %s, need %s", current_amount, needed_amount
            )
            _update_progress(
                self.progress,
                100 * current_amount / needed_amount,
                f"Capturing data {current_amount}/{needed_amount}",
            )
            try:
                next_event = self._next_validated_event_or_raise(
                    daq_buffer, self._timeout, self._attempts
                )
            except PedestalsDataCaptureError:
                self._stop_capture()
                raise
            except OperationCanceledError:
                LOGGER.debug("Pedestals generation canceled")
                break

            valid_events.append(next_event)

            # How long we keep reading for depends on whether we have enough data
            # for ALL windows. `current_amount` is the lowest capture amount across all windows.
            current_amount = np.min(self._get_window_counts(valid_events)[channels])

        self._stop_capture()

        self.validated_data = valid_events

    def _next_validated_event_or_raise(
        self, buffer, timeout: float, attempts: int
    ) -> dict:
        """Guaranteed to either return a valid event, or raise an error.
        This method must be called while capturing.

        Args:
            buffer (deque): daq output buffer
            timeout (float): expected timeout in seconds
            attempts (int): number of attempts

        Returns:
            dict: the event.

        Raises:
            PedestalsDataCaptureError: if a valid event could not be read.
            OperationCanceledError: if pedestals generation was canceled
        """
        event = None
        for i in range(attempts):
            self._raise_if_canceled()  # exit point for canceling
            self._toggle_trigger_or_raise()

            try:
                self._wait_for_event_or_raise(buffer, timeout)
                event = buffer.popleft()
            except (TimeoutError, IndexError):
                LOGGER.warning("Timed out while waiting for an event")
                get_board_controller(self.board).stop_readout()
                time.sleep(0.1)
                get_board_controller(self.board).digital_reset()
                time.sleep(0.25)
                get_board_controller(self.board).digital_reset()
                time.sleep(0.1)
                get_board_controller(self.board).set_continuous_mode(True)
                time.sleep(0.05)
                get_board_controller(self.board).start_readout("software")
                continue

            if self._validate_event(event):
                LOGGER.debug("Got a validated event")
                time.sleep(0.5)  # Let the trigger reset
                break
        else:
            msg = (
                "Failed to capture an event. The board may be unresponsive "
                "or need to be power cycled/reinitialized."
            )
            LOGGER.error("Failed to capture an event!")
            raise PedestalsDataCaptureError(msg)

        return event

    def _wait_for_event_or_raise(self, buffer, timeout):
        """Wait for an event or raise an error. If the function
        returns it is guaranteed that an event was read out.

        Args:
            buffer (Iterable): the daq output buffer
            timeout (float): timeout

        Raises:
            PedestalsDataCaptureError: if the event cannot be read in time.
        """
        waiter = EventWaiter(buffer, amount=len(buffer) + 1, timeout=timeout)
        try:
            waiter.start(blocking=True)
        except TimeoutError:
            raise
        except Exception as e:
            msg = "Pedestals data capture failed due to unknown reason"
            raise PedestalsDataCaptureError(msg) from e

    def _toggle_trigger_or_raise(self):
        """Toggle the external trigger or raise ``PedestalsDataCaptureError`` on failure"""
        try:
            get_board_controller(self.board).toggle_trigger()
        except Exception as e:
            msg = (
                "Pedestals data capture failed: software trigger could not "
                "be sent. Is the board connected?"
            )
            raise PedestalsDataCaptureError(msg) from e

    def _start_capture(self):
        """Start capturing data"""
        self.board.connection.reset_input_buffer()
        self._daq.output_buffer = deque()
        self._daq.start_capture()
        time.sleep(0.1)
        self._board_controller.set_continuous_mode(True)
        self._board_controller.start_readout("software")

    def _stop_capture(self):
        """Stop capturing data"""
        self._board_controller.stop_readout()
        self._daq.stop_capture()
        self._daq.stop_workers()

    # Helpers/reimplemented =========================================================================
    def _backup_settings(self) -> dict:
        """Store relevant registers"""
        return {}

    def _restore_backup_settings(self, backup):
        """Restore backed up registers"""
        pass

    def _validate_event(self, event):
        """Check if the event has a data field, which means it's parsed"""
        if "data" not in event:
            return False
        chans_with_data = [i for i, x in enumerate(event.get("data", [])) if len(x) > 0]
        return set(chans_with_data).issuperset(self._channels)

    def _get_window_counts(self, events: list[dict]) -> np.ndarray:
        """Calculate the number of times each window occurs in the given events.

        Args:
            events (list[dict]): list of validated events.

        Returns:
            np.ndarray: 2D int array with shape (channels, windows) containin
                window counts.
        """
        channels = self.board.channels
        windows = self.board.params["windows"]
        window_hits = np.zeros((channels, windows), dtype=int)

        # Count the number of times each window shows up in the data.
        for event in events:
            for chan, chan_window_labels in enumerate(event["window_labels"]):
                if len(chan_window_labels) == 0:
                    continue
                # skip first window, it's junk.
                window_hits[chan][chan_window_labels[1:-1]] += 1
        return window_hits

    # Data generation =========================================================================
    def _generate_pedestals_data(self, validated_data=None, num_captures=None):
        """Assemble the data from a list of validated events into the pedestals raw data array.
        The name of this function is extremely misleading!

        Args:
            validated_data (list[dict]): list of validated events
            num_captures (int): number of captures
        """
        num_captures = num_captures or self.num_captures
        validated_data = validated_data or self.validated_data
        num_warmup = self.num_warmup_events

        LOGGER.debug("Generating pedestals from %s samples", num_captures)
        _update_progress(self._progress, 95, "Generating pedestals...")
        channels = self.board.params.get("channels", 96)
        windows = self.board.params.get("windows", 64)
        samples = self.board.params.get("samples", 64)
        rawdata = self.board.pedestals["rawdata"]
        window_counts = np.zeros((channels, windows), dtype=int)

        for event in validated_data:
            for chan in range(channels):
                chan_data = event["data"][chan]
                for window_idx, window in enumerate(event["window_labels"][chan]):
                    if (
                        window_idx == 0 or window_idx == windows - 1
                    ):  # first window is junk
                        continue

                    cap = window_counts[chan, window] - num_warmup
                    window_counts[chan, window] += 1
                    if cap < 0 or cap >= num_captures:  # skip first few captures
                        continue

                    data = chan_data[window_idx * samples : (window_idx + 1) * samples]
                    rawdata[chan, window, :, cap] = data

        return True

    def _generate_pedestals_from_data(self):
        """Generate pedestals data and average it.

        Takes the generated pedestals data and generates the actual pedestals.
        The pedestals are stored in the controller and can be retreived with
        get_pedestals().

        """
        pedestals = self.board.pedestals
        rawdata = pedestals["rawdata"]
        output = np.nanmean(rawdata, axis=3)
        np.nan_to_num(output, nan=0, copy=False)
        pedestals["data"] = output



"""Answer parser.

"""

from logging import getLogger
from threading import Thread

from naludaq.parsers.answer_parser import get_answer_parser

logger = getLogger(__name__)


class AnswerParserWorker(Thread):
    """Answer parser separates the answer and question id from raw data."""

    def __init__(self, input_buffer, output_buffer: dict, lock):
        super().__init__()
        self.running = True
        self.daemon = True
        self.input_buffer = input_buffer
        self.answers = output_buffer
        self.lock = lock
        self.answer_parser = get_answer_parser(True)

    def run(self):
        logger.debug("Trying to start packager.")
        self._worker_loop()

    def stop(self):
        self.running = False

    def _worker_loop(self):
        """Parse answers in the input buffer.

        Waits until an answer is available in the buffer.

        The parsed answers have a specific format.
        HHVVIIEE:
        HH - header
        VV - value
        II - command unique id
        EE - end word

        """
        answer_parser = self.answer_parser
        while self.running:
            buff = b""
            try:
                buff = self.input_buffer.get(block=True, timeout=1)

            except:
                # logger.error("answer_parser buff: %s", buff)
                continue
            else:
                buff_data = buff["rawdata"]

            if len(buff_data) != 8:
                logger.error(
                    "DAQ.answer_parser couldn't parse answer due  length of package, expected 8 got: %s",
                    len(buff_data),
                )
                continue

            try:
                answer = answer_parser(buff_data)

            except Exception as error_msg:
                logger.error(
                    "DAQ.answer_parser couldn't unpack answer due to %s", error_msg
                )
                continue

            with self.lock:
                logger.error(
                    "Fond a response: %s with id: %s", answer["value"], answer["cmd_id"]
                )
                self.answers[answer["cmd_id"]] = answer["value"]



from .helper_functions import *

__all__ = [
    "create_dir_if_not_exist",
    "get_application_path",
    "locked_deque",
    "hex_to_int",
    "int_to_bin",
    "swap_dict_key_and_val",
    "type_name",
]



"""Convert ADC 2 mV.

Tool to convert the data fields in an event dict from ADC counts to mV.
It requires the board to have calibration data loaded as an attribute `caldata`


"""
# import operator
# import pathlib
from copy import deepcopy

import numpy as np

# from naludaq.board import Board

# from nalu_analyzer.helper_functions import open_data


class ADC2mVConverter:
    def __init__(self, params: dict = None, caldata: dict = None):
        self._caldata = None
        self.lr_dict = dict()
        self.params = params
        self.caldata = caldata

    @property
    def caldata(self):
        """Get/Set the calibration data.

        {
            "slope": 2D Array, channel major, sample minor,
            "intercept": 2D Array, channel major, sample minor,
            "linear_region": list of minmax tuples per channel
        }
        """
        return self._caldata

    @caldata.setter
    def caldata(self, caldata):
        self._caldata = caldata

    def run(self, event: dict, immutable: bool = True, lin_shift: bool = False) -> dict:
        """Convert the data field from adc counts2 mV.

        Args:
            evt: event to correct, must contain a data field
            immutable: Overwrite original or create a copy
            lin_shift: linear correction constants

        Returns:
            converted event or original event if there's no data field.
        """
        if event.get("data", None) is not None:
            return_evt = self._adc2mv_sample(event, immutable, lin_shift)
        else:
            return_evt = event
        return return_evt

    def _adc2mv_sample(
        self, evt: dict, immutable: bool = True, lin_shift: bool = False
    ):
        """Converts ADC to mV using correction constants for each sample.

        This function iterates of each individial sample and correct it.
        The function can either overwrite the original event or create a copy.
        Note the copy operation is time consuming.

        Args:
            evt: event to correct, must contain a data field
            immutable: Overwrite original or create a copy
            lin_shift: linear correction constants

        Returns:
            New converted event if immutable is True, else converts the input event.
        """
        evt_out = evt
        if immutable:
            data_mv_evt = deepcopy(evt)
            evt_out = data_mv_evt

        converted = list()

        window_labels = evt["window_labels"]

        samples = self.params["samples"]
        channels = self.params["channels"]

        for chan in range(channels):

            winds = window_labels[chan]
            num_winds = len(winds)
            if num_winds == 0:
                converted.append(np.array([], dtype="float"))
                continue

            locs = (samples * np.repeat(winds, samples)) + (
                np.ones((num_winds, samples), dtype=int)
                * np.arange(0, samples, dtype=int)
            ).flatten()

            try:
                if lin_shift:
                    converted.append(
                        self._inv_linear_shift(
                            evt_out["data"][chan],
                            slope=self.caldata["slope"][chan],
                            intercept=self.caldata["intercept"][chan],
                        )
                    )

                else:
                    converted.append(
                        self._inv_linear_function(
                            evt["data"][chan],
                            slope=self.caldata["slope"][chan].take(locs, mode="wrap"),
                            intercept=self.caldata["intercept"][chan].take(
                                locs, mode="wrap"
                            ),
                        )
                    )
            except:
                raise

        evt_out["data"] = np.array(converted)
        return evt_out

    def _linear_function(self, x_data, slope, intercept):
        """
        This function will apply a linear transformation on the x_data, in the form:
        return_data = (slope * x_data) + intercept.
        """
        return (np.array(slope) * x_data) + intercept

    def _inv_linear_function(self, x_data, slope, intercept):
        """
        This function will apply a inv. linear transformation on the x_data, in the form:
        return_data = (x_data - intercept) / slope
        """
        return (np.array(x_data) - intercept) / slope

    def _inv_linear_shift(self, x_data, slope):

        """
        This function will apply a inv. linear transformation on the x_data, in the form:
        return_data = (x_data - intercept) / slope
        """
        return (np.array(x_data)) / slope



import itertools
import time
from collections import deque

import numpy as np

from naludaq.backend import AcquisitionManager, ConnectionManager
from naludaq.communication import AnalogRegisters
from naludaq.controllers import (
    get_board_controller,
    get_dac_controller,
    get_readout_controller,
)
from naludaq.daq.workers.packager import get_packager
from naludaq.daq.workers.worker_serial_reader import SerialReader
from naludaq.helpers import event_transfer_time
from naludaq.helpers.exceptions import BadDataError
from naludaq.parsers import get_parser
from naludaq.tools import EventWaiter

from .bayesian_optimizer import BayesianOptimizer


class GainStageTuner(BayesianOptimizer):
    def __init__(self, board, bounds, score_function=None):

        if score_function is None:
            score_function = self.calculate_mean_diff_from_acq

        self.score_function = score_function
        self.cost_function = self.score_variable_pair
        super().__init__(board, self.cost_function, bounds)
        self.board = board
        self.bounds = bounds

        self.events_per_probe = 10
        self.readout_params = {"trig": "imm", "lb": "forced"}

        self.event_params = {
            "windows": 10,
            "lookback": 20,
            "write_after_trig": 254,
        }

    def run(self, n_iter=50):
        results = self.maximize(n_iter)

        suggested_isel = round(results["params"]["y"])
        suggested_dac = round(results["params"]["x"])

        AnalogRegisters(self.board).write("isel", suggested_isel)
        get_dac_controller(self.board).set_dacs(suggested_dac)

    def score_variable_pair(self, x: int, y: int) -> float:
        """Sets the ISEL and DAC value to a certain value on the board, then captures some pedestals
        data and scores the data capture based on how close the average value is to midrange. The
        higher the value, the closer all 4 channels are to being averaged around 2048.

        Args:
            x (int): DAC Value. HAS to be named x since that is the kwarg name that bayes_opt module uses
            y (int): ISEL Value. HAS to be named y since that is the kwarg name that bayes_opt module uses

        Returns:
            float: Score based on how close all channel averages are to midrange. Higher the number,
                closer the value is to midrange
        """
        x = round(x)
        y = round(y)

        # Apply parameters
        AnalogRegisters(self.board).write("isel", y)
        get_dac_controller(self.board).set_dacs(x)

        # Get Data
        acq = self._get_data()

        # Calculate Score
        score = self.score_function(acq)

        return score

    def _get_data(self):

        data = self._protected_readout()
        return data

    def _readout_packages(self, timeout: int = 5) -> deque:

        if self.board.using_new_backend:
            return self._backend_readout(timeout)

        return self._classic_readout(timeout)

    def _classic_readout(self, timeout: int = 5) -> deque:
        """Basic readout function based on naludaq's example script for acquiring waveform data.

        Args:
            board (Board): board object with active connection
            trig (str, optional): trigger type. Defaults to 'imm'.
            lb (str, optional): lookback type. Defaults to 'forced'.
            timeout (int, optional): max amt of time to wait for evt data (seconds). Defaults to 5.

        Returns:
            deque: deque of raw captured waveform data, in bytearrays
        """

        serial_buffer = deque()
        output_buffer = deque()
        evt_waiter = EventWaiter(output_buffer, self.events_per_probe, timeout=timeout)

        stopword = self.board.params["stop_word"]
        sr = SerialReader(self.board.connection, serial_buffer)
        pk = get_packager(
            self.board, serial_buffer, output_buffer, deque(), stopword, 100
        )

        while self.board.connection.in_waiting:
            self.board.connection.reset_input_buffer()

        _bc = get_board_controller(self.board)

        _bc.start_readout(
            trig=self.readout_params["trig"],
            lb=self.readout_params["lb"],
            singleEv=True,
        )
        pk.start()
        sr.start()

        evt_waiter.start()

        sr.stop()
        pk.stop()
        _bc.stop_readout()

        return output_buffer

    def _backend_readout(self, timeout: int = 5) -> deque:
        """
        Backend readout function for acquiring waveform data using NaluDaq's Rust Backend.
        Incompatible with the old way of reading out (anything that has "get_x_connection")

        Args:
            timeout (int, optional): max amt of time to wait for evt data (seconds). Defaults to 5.

        Returns:
            deque: deque of raw captured waveform data, in bytearrays
        """

        output_buffer = deque()
        timeout = timeout / self.events_per_probe

        _bc = get_board_controller(self.board)

        # Clears the I/O buffers for whatever connection thats being used
        ConnectionManager(self.board).device.clear_buffers()

        _bc.start_readout(
            trig=self.readout_params["trig"],
            lb=self.readout_params["lb"],
            singleEv=True,
        )

        # use a temporary acquisition to stream data into an output buffer
        try:
            with AcquisitionManager(self.board).create_temporary() as acq:
                acq.set_output()
                stream = acq.stream(timeout=timeout)

                output_buffer.extend(itertools.islice(stream, self.events_per_probe))

        except TimeoutError:
            raise BadDataError

        _bc.stop_readout()

        return output_buffer

    def _configure_readout(self) -> None:
        """Prepares the board with the proper readout parameters. To be used BEFORE readout.

        Args:
            board (board): board with active connection
            windows (int): number of windows to capture per trigger
            lookback (int): pre-trigger length (in windows)
            write_after_trig (int): record length (in windows)
            events (int): number of trigger events to capture

        """
        _rc = get_readout_controller(self.board)
        _rc.number_events_to_read(self.events_per_probe)
        _rc.set_readout_channels(list(range(self.board.params["channels"])))
        _rc.set_read_window(**self.event_params)

    def _protected_readout(self) -> list:
        """I needed a function that is protected against BadDataError raises, which occur presumably
        because my wait time is too short. I don't know the exact reason, but trying again over and over
        seems to work so I'm sticking with it for now.

        Args:
            board (Board): board object with active connection
            trig_type (str): trigger type of acquisition
            lb_type (str): lookback type of acquisition
            events (int): number of events to capture (1 trigger = 1 event)

        Returns:
            pdata (list): list of parsed events
        """

        self._configure_readout()

        pdata = []
        fails = 0
        delay_per_evt = event_transfer_time(
            self.board, windows=self.event_params["windows"], overhead=0.5
        )
        delay = delay_per_evt * self.events_per_probe

        parser = get_parser(self.board.params)

        while len(pdata) == 0 and fails < 5:
            try:
                data = self._readout_packages(timeout=delay)
                pdata = [parser.parse(evt) for evt in data]
                _ = [self._verify_evt_or_raise(evt) for evt in pdata]
            except BadDataError:
                print(f"Bad Data Error: Fails: {fails}")
                time.sleep(1)  # give the board some time to get its shit together
                fails += 1

        if fails >= 5:
            print("Tried 5 times. Failed. Giving Up")
            return 0

        return pdata

    def calculate_mean_diff_from_acq(self, acq):
        """Calculates the difference from the acq's average to midrange (2048)."""

        evt_averages = []

        for evt in acq:
            evt_averages.append([np.mean(chandata) for chandata in evt["data"]])

        chan_averages = np.mean(evt_averages, axis=0)

        # compute score (also known as the cost function)
        cost_value = 0 - np.sum([np.abs(avg - 2048) for avg in chan_averages])

        return cost_value

    def _verify_evt_or_raise(self, evt):

        if type(evt) is not dict:
            raise BadDataError("evt is not dictionary")

        if "data" not in evt.keys():
            raise BadDataError("evt does not have 'data' field")

        if evt["data"].dtype not in ["uint16", "uint32"]:
            raise BadDataError(f"evt data is wrong type: {evt['data'].dtype}")



from naludaq.communication.registers import Registers


class RegisterCache:
    def __init__(self, board) -> None:
        """Cache utility for software registers.

        Args:
            board (Board): the board object, the source of register values.
        """
        self._board = board
        self._cache = {}

    @property
    def board(self):
        return self._board

    def add(self, name: str, reg_type: Registers):
        """Adds a register to be tracked by the cache.

        Args:
            name (str): the register name
            reg_type (Registers): the type of register (e.g. AnalogRegisters).

        Raises:
            TypeError if arguments are an invalid type
            KeyError if the register is already added, or is not a valid register.
        """
        if not isinstance(name, str):
            raise TypeError(f"Name must be a string, not {type(name).__name__}")
        try:
            if not issubclass(reg_type, Registers):
                raise TypeError("Register type must be a subclass of Registers")
        except:
            raise TypeError("Registers must be a class")
        if name in self._cache:
            raise KeyError(f'The register "{name}" is already added.')
        try:
            value = reg_type(self._board).registers[name]["value"]
        except KeyError:
            raise KeyError(f'Register "{name}" is not a valid register')
        self._cache[name] = [reg_type, value]

    def get(self, name: str) -> int:
        """Gets a register value from the cache.

        Args:
            name (str): the register name.

        Returns:
            The value of the register

        Raises:
            TypeError if arguments are an invalid type
            KeyError if the register has not been added to the cache.
        """
        if not isinstance(name, str):
            raise TypeError(f"Name must be a string, not {type(name).__name__}")
        try:
            return self._cache[name][1]
        except KeyError:
            raise KeyError(f'The register "{name}" has not been added.')

    def update(self, name: str, value: int = None):
        """Updates the value in the cache.

        Args:
            name (str): the name of the register.
            value (int): the value to set the register to.
                If `None`, the value from the software
                registers is used instead.

        Raises:
            TypeError if arguments are an invalid type
            KeyError if the register has not been added.
        """
        if not isinstance(name, str):
            raise TypeError(f"Name must be a string, not {type(name).__name__}")
        if not isinstance(value, (int, type(None))):
            raise TypeError("Value must be int or None")
        try:
            if value is None:
                value = self._get_software_register(name)
            self._cache[name][1] = value
        except KeyError:
            raise KeyError(f'The register "{name}" has not been added.')

    def update_all(self, reg_values: dict = None):
        """Updates the cached values for all (or a subset of)
        all tracked registers.

        Args:
            reg_values (dict): a dict of {reg_name: reg_value} to
                update the cache with. If `None`, all registers are
                updated with the software registers.

        Raises:
            TypeError if arguments are an invalid type
            KeyError if one or more registers are not added.
        """
        if not isinstance(reg_values, (dict, type(None))):
            raise TypeError(
                f"Register values must be a dict or None, not {type(reg_values).__name__}"
            )
        if reg_values is None:
            for name in self._cache:
                self.update(name)
        else:
            for name, value in reg_values.items():
                self.update(name, value)

    def restore(self, name: str):
        """Writes the value held in the cache to the board.

        Args:
            name (str): the reigster name.

        Raises:
            TypeError if arguments are an invalid type
            KeyError if the register is not added.
        """
        if not isinstance(name, str):
            raise TypeError(f"Name must be a string, not {type(name).__name__}")
        try:
            value = self.get(name)
            self._get_register_obj(name).write(name, value)
        except KeyError:
            raise KeyError(f'The register "{name}" has not been added.')

    def restore_all(self):
        """Writes all values held in the cache to the board.

        Raises:
            KeyError if the register is not added.
        """
        for name in self._cache:
            self.restore(name)

    def _get_register_obj(self, name: str) -> Registers:
        """Gets an instantiated `Registers` object that
        can be used to communicate with the given register.

        Args:
            name (str): the name of the register.

        Returns:
            The `Registers` object.

        Raises:
            TypeError if arguments are an invalid type
            KeyError if the register has not been added
        """
        if not isinstance(name, str):
            raise TypeError(f"Name must be a string, not {type(name).__name__}")
        try:
            return self._cache[name][0](self._board)
        except KeyError:
            raise KeyError(f'The register "{name}" has not been added.')

    def _get_software_register(self, name: str):
        """Gets the value of a software register.

        Args:
            name (str): the register name

        Returns:
            The software register value

        Raises:
            TypeError if arguments are an invalid type
            KeyError if the register has not been added
        """
        if not isinstance(name, str):
            raise TypeError(f"Name must be a string, not {type(name).__name__}")
        try:
            return self._get_register_obj(name).registers[name]["value"]
        except KeyError:
            raise KeyError(f'The register "{name}" has not been added.')



"""Helper functions for validating inputs."""
import pathlib


def validate_dir_or_raise(output_dir, name="Directory"):
    if not isinstance(output_dir, (str, pathlib.Path)):
        raise TypeError(f"{name} must be a string.")
    if not pathlib.Path(output_dir).exists():
        raise FileNotFoundError(f"{name} must exist.")
    if output_dir is None or not pathlib.Path(output_dir).is_dir():
        raise NotADirectoryError(f"{name} must be specified.")



"""
"""
import time
from logging import getLogger

from naludaq.backend.managers.connection import ConnectionManager
from naludaq.backend.models.device import DeviceType
from naludaq.communication import ControlRegisters
from naludaq.controllers.controller import Controller
from naludaq.helpers.helper_functions import type_name

LOGGER = getLogger("naludaq.board_controller_upac96")


class UPAC96BoardController(Controller):
    """Board controller for the specific UPAC96 functionality."""

    def __init__(self, board):
        super().__init__(board)

    def start_readout(self, trigger_mode: str):
        """Start a readout.

        Args:
            trigger_mode (str): the trigger mode to set as specified
                in the documentation for ``set_trigger_mode``. If the
                argument is not provided the trigger type is not set.
        """
        self.set_trigger_mode(trigger_mode)
        self.arm_board()

    def stop_readout(self):
        """Not sure if this is the way the board operates anymore, but this should stop the readout."""
        self.set_trigger_mode("software")
        self.set_continuous_mode(False)

    def set_trigger_mode(self, mode: str):
        """Select the trigger mode for the board.

        Possible trigger modes are:
        - 'software' (power-on default)
        - 'self' (UDC self trigger)
        - 'auto' (FPGA auto trigger 1 Hz)
        - 'external' (external trigger)

        Args:
            mode (str): the trigger mode string as specified above, or
                a portion of it (e.g. 'ext' is interpreted as 'external')

        Raises:
            ValueError: if the mode string is ambiguous ('s').
        """
        if not isinstance(mode, str):
            raise TypeError(f"Trigger mode must be str, not {type_name(mode)}")

        available_modes = {
            "software": 0b00,
            "self": 0b01,
            "auto": 0b10,
            "external": 0b11,
        }

        # An 's' string is ambiguous...
        mode = mode.lower()
        mode_values = [
            (long_name, value)
            for long_name, value in available_modes.items()
            if long_name.startswith(mode)
        ]
        if len(mode_values) == 0:
            raise ValueError(f'Invalid trigger mode "{mode}"')
        if len(mode_values) > 1:
            disambiguation = [x[0] for x in mode_values]
            raise ValueError(
                f'Trigger mode "{mode}" is ambiguous. It could mean '
                f'{", ".join(disambiguation[:-1])}, or {disambiguation[-1]}. '
                f"Please provide a longer string."
            )

        name, value = mode_values[0]
        LOGGER.debug("Setting trigger mode to %s (%s)", value, name)
        self._write_control_register("trigger_select", value)

    def arm_board(self):
        """Arm the board for readout."""
        LOGGER.debug("Arming board")
        self._write_control_register("arm", 1)
        self._write_control_register("arm", 0)

    def set_continuous_mode(self, continuous: bool):
        """Turns on/off continuous mode.

        If continuous mode is turned off (default) while in software mode,
        the board needs to be rearmed after every trigger event.
        """
        if continuous not in [True, False]:  # also allows 0 or 1
            raise TypeError(f"Argument must be bool, not {type_name(continuous)}")
        LOGGER.debug("Set continuous mode: %s", continuous)
        self._write_control_register("continuousmode", continuous)

    def toggle_trigger(self, cycles: int = 3):
        """Toggle the trigger signal."""
        if not isinstance(cycles, int):
            raise TypeError(f'"cycles" must be an int, not {type_name(cycles)}')
        if cycles < 1:
            raise ValueError(f'"cycles" must be at least 1, got {cycles}')
        if cycles > 2**16 - 1:
            raise ValueError(
                f'"cycles" must be at most {2**16 - 1} (16-bits), got {cycles}'
            )
        self.arm_board()

        LOGGER.debug("Sending software trigger")
        cmd = f"C000{cycles:04X}"
        self.board.connection.send(cmd)

    def toggle_reread(self):
        """Re-read the last event.

        This will disable continuous mode.
        """
        LOGGER.debug("Toggling reread")

        # Reread cannot start in continuous mode
        cr = ControlRegisters(self.board)
        prev_continuous_mode = cr.registers["continuousmode"]["value"]
        self.set_continuous_mode(False)

        # Sleep time allows the state machine to return to "IDLE" state
        # after disabling continuous mode. Makes re-read waaay more stable.
        time.sleep(0.05)
        self._write_control_register("reread_data", True)
        self._write_control_register("reread_data", False)
        time.sleep(0.01)
        self.set_continuous_mode(prev_continuous_mode)

    def clear_buffer(self):
        """Clears the UART buffer on both CPU and FPGA side."""
        self._clear_fpga_buffer()
        self._clear_uart_buffer()

    def _clear_uart_buffer(self):
        """Clears the UART buffer on the"""
        if self.board.using_new_backend:
            device = ConnectionManager(self.board).device
            if device.type in [DeviceType.SERIAL, DeviceType.D2XX]:
                device.clear_buffers()
        else:
            self.board.connection.reset_input_buffer()

    def _clear_fpga_buffer(self):
        """Clear the FPGA FIFOs for both UART and USB data"""
        original_usb = self.board.registers["control_registers"]["usb_fifo_disable"][
            "value"
        ]
        original_uart = self.board.registers["control_registers"]["uart_fifo_disable"][
            "value"
        ]
        self._write_control_register("usb_fifo_disable", 1)
        self._write_control_register("uart_fifo_disable", 1)
        time.sleep(0.2)
        self._write_control_register("usb_fifo_disable", original_usb)
        self._write_control_register("uart_fifo_disable", original_uart)
        time.sleep(0.2)

    def enable_usb(self):
        """Enable the USB interface on the board while disabling the UART interface."""

        usb = True
        uart = False
        self.enable_communication_interfaces(usb, uart)

    def enable_uart(self):
        """Enable the UART interface on the board while disabling the USB interface."""
        usb = False
        uart = True
        self.enable_communication_interfaces(usb, uart)

    def enable_communication_interfaces(self, usb: bool, uart: bool):
        """Independently control the USB and UART interfaces.

        Args:
            usb (bool): enable/disable the USB interface
            uart (bool): enable/disable the UART interface

        Raises:
            TypeError: if the arguments are not bools.
        """
        if not isinstance(usb, bool):
            raise TypeError(f"Argument usb must be bool, not {type_name(usb)}")
        if not isinstance(uart, bool):
            raise TypeError(f"Argument uart must be bool, not {type_name(uart)}")

        self._write_control_register("usb_fifo_disable", True)
        self._write_control_register("uart_fifo_disable", True)
        time.sleep(0.2)
        self._write_control_register("usb_fifo_disable", not usb)
        self._write_control_register("uart_fifo_disable", not uart)
        time.sleep(0.2)

    def power_on_rails(self, rail1: bool = True, rail2: bool = True):
        """Power on the chips on the board.

        Args:
            rail1 enables chips 0-2,
            rail2 enables chips 3-5.
        """
        self._write_control_register("pwr_udc1_en", rail1)
        self._write_control_register("pwr_udc2_en", rail2)

    def reset_board(self):
        """Try and reset the board.

        In case the FPGA get stuck this can help reset the state.
        """
        time.sleep(0.02)
        self.digital_reset()
        time.sleep(0.02)
        self.sysrst()

    def digital_reset(self):
        """Toggles the "reset" port on the readout module.

        Forcibly returns the chip to default state.
        """
        self._write_control_register("digrst", True)
        self._write_control_register("digrst", False)

    def sysrst(self):
        """Toggles the sysrst pin, which resets the digital portion of the chip"""
        self._write_control_register("sysrst", True)
        self._write_control_register("sysrst", False)

    def get_available_chips(self) -> list[int]:
        """Get a list of available chip numbers."""
        lock_bits = self._read_control_register("udc_rxout_locked")
        num_chips = self.board.params.get("num_chips", 6)
        chips = [c for c in range(num_chips) if (lock_bits >> c) & 1]
        return chips

    def set_trigger_monitoring_disabled(self, disabled: bool):
        """Enable monitoring of the trigger signal."""
        self._write_control_register("trigger_monitor_disable", disabled)

    def _read_control_register(self, register: str) -> int:
        """Reads a control register"""
        return ControlRegisters(self.board).read(register)["value"]

    def _write_control_register(self, register, value):
        """Writes a control register"""
        ControlRegisters(self.board).write(register, value)  # pragma: no cover



"""Parser

Use parser to parse the data from the Hardware.
The boards sends a raw bitstream and it needs to be parsed into events.
The parser class is a tool to parse the raw bitstream of 8-bit chunks
into 16 bit words then extracting the 12-bit data, the header info,
channels and window number data.

This data is then returned in a dictionary.
The raw data is preserved.

This module was previously located under the naludaq.daq.workers
subpackage, and was relocated to improve the organization of the
naludaq package.
"""

import logging

import numpy as np

from naludaq.helpers.exceptions import BadDataError, IncompatibleParserError
from naludaq.parsers.headers import get_header_parser

LOGGER = logging.getLogger(__name__)


class Parser:
    """Class containing the parsers.

    Functions:
    ----------
        parse: Parse the raw event based on set parameters.
        parse_digital_data: raw event to a dict, no time.

    Args:
        params (dict): Contains hardware parameters.
        minimum_params = {
            'channels': amount of channels,
            'samples': amount of samples,
            'chanmask': bits for channels in the window headers,
            'windmask': bits for windows number in the window headers,
            'chanshift': bits the channel number is shifted up in windows headers,
            'headers': expected amount of event headers,
        }

    Attributes:
        num_wind_headers (int): Number of window  headers
        num_footers (int): Number of bytes in the footer
        last_bits (int): how many bits in the end of the word.
        data_bitmask (int): bits of data before shifting last bit
    """

    def __init__(self, params):
        self.params = params
        self.samples = params["samples"]
        self.channels = params["channels"]
        self.windows = params["windows"]
        self._parse_event_headers = get_header_parser(self.params["model"])
        self.num_wind_header = 1
        self.num_footers = 2
        self.last_bits = self.params.get("lastbits", 1)
        self.data_bitmask = 8191
        self.validation_patterns = None
        # Default to old behaviour.
        if self.params["model"] == "aardvarcv3":
            raise IncompatibleParserError(
                "This parser is not compatible with AARDVARCv3 data."
            )
        elif self.params.get("use_new", False) is True:
            self.parse_digital_data = self.parse_digital_data_new
        else:
            self.parse_digital_data = self.parse_digital_data_old

    def parse(self, raw_data) -> dict:
        """Parses raw_data into a dict.

        Parses the data from the raw event and strips the timestamp/id.
        An dict is created from the parsed data with the id preserved.

        Args:
            raw_data (bytes): Raw data packaged with a header and a stopword.
        Returns:
            dict.
        """
        event = {}
        try:
            event = self.parse_digital_data(raw_data)
        except (TypeError, Exception) as e_msg:
            LOGGER.exception("parse_digital_data failed: %s", e_msg)
            event = raw_data
            # raise Exception(f'parse_digital_data failed: {e_msg}')
        else:
            # Add the X-Axis.
            event["time"] = self._add_xaxis_to_event(event)

            event["created_at"] = raw_data.get("created_at", 0)
            event["pkg_num"] = raw_data.get("pkg_num", 0)
            event["event_num"] = raw_data.get("pkg_num", 0)

        event["name"] = None

        return event

    def parse_digital_data_new(self, in_data) -> dict:
        """Parse the raw data from the board.

        Parses the data spit out by the digitial readout portion of the firmware

        Args:
            in_data (bytearray): Raw data from the board
        Returns:
            Parsed event as a dict.
        Raises:
            BadDataException if no data is found or if the data contains errors.

        Additional info:
        The word is 16-bit and the data is 12-bit.
        0x00 is not a valid byte, ever

        Example format for ASoC:
        first three bits are the "identifier", last bit is always 1
        4 x event headers:
          '001' 'prev_final_window' '1'
          '001' 'trigger_time_ns(23 dt 12)' '1'
          '001' 'trigger_time_ns(11 dt 0 )' '1'
          '001' 'eventNumber' '1'
        { 1x *digitization header*:
          "001" '12-bit timer' '1'
         [ 1x window header:
          '100' 'chan' '00' 'window' '1'
          64x data packets
           '100' 'data' '1'
           ] for every channel
          } for every window
        1x end packet (x"FACE")

        """
        # try:
        #    self._validate_input_package(in_data)
        # except:
        #    raise
        # Convert the 8-bit binary to 16-bit unsigned int and bit mask+shift the entire array
        raw_data = np.frombuffer(in_data["rawdata"], dtype=">H")
        bitshifted_data = self._bitshift_data(raw_data)
        curr_event = {
            "window_labels": [[] for _ in range(self.params["channels"])],
            "data": [[] for _ in range(self.params["channels"])],
            "start_window": (raw_data[self.params["headers"]] & self.params["windmask"])
            >> 1,
            "digitize_times": list(),
        }

        self._parse_event_headers(curr_event, bitshifted_data)

        # Determine how many digitization headers there are
        num_headers = self.params["headers"]
        identifiers = raw_data >> 13
        dig_head_locs = np.where(identifiers == 1)[0][num_headers:]
        num_dig_headers = len(dig_head_locs)
        curr_event["num_dig_headers"] = num_dig_headers
        curr_event["digitize_times"] = bitshifted_data[dig_head_locs]
        LOGGER.debug("dig_times:%s", curr_event["digitize_times"])

        # Move through data one window at the time.
        data_locs = np.where(identifiers == 4)[0]
        LOGGER.debug("data_locs:%s", data_locs)
        step_size = self.num_wind_header + self.params["samples"]
        for window_idx in data_locs[::step_size]:

            # Extract the window header data
            channel = (raw_data[window_idx] & self.params["chanmask"]) >> self.params[
                "chanshift"
            ]
            window = (raw_data[window_idx] & self.params["windmask"]) >> self.last_bits
            curr_event["window_labels"][channel].append(window)

            curr_event["data"][channel].extend(
                bitshifted_data[
                    window_idx + 1 : window_idx + (self.params["samples"] + 1)
                ]
            )

        curr_event["data"] = np.array(curr_event["data"])

        return curr_event

    def parse_digital_data_old(self, in_data) -> dict:
        """Parse the raw data from the board.

        Parses the data spit out by the digitial readout portion of the firmware

        Args:
            in_data (bytearray): Raw data from the board
        Returns:
            Parsed event as a dict.
        Raises:
            BadDataException if no data is found or if the data contains errors.

        Additional info:
        The word is 16-bit and the data is 12-bit.
        0x00 is not a valid byte, ever

        Example format for ASoC:
        first three bits are the "identifier", last bit is always 1
        4 x event headers:
          '001' 'prev_final_window' '1'
          '001' 'trigger_time_ns(23 dt 12)' '1'
          '001' 'trigger_time_ns(11 dt 0 )' '1'
          '001' 'eventNumber' '1'
        1x window header:
          '010' 'chan' '00' 'window' '1'
        64x data packets
          '100' 'data' '1'
        1x end packet (x"FACE")

        """
        try:
            self._validate_input_package(in_data)
        except:
            raise
        # Convert the 8-bit binary to 16-bit unsigned int and bit mask+shift the entire array
        raw_data = np.frombuffer(in_data["rawdata"], dtype=">H")
        bitshifted_data = self._bitshift_data(raw_data)

        curr_event = {
            "window_labels": [[] for _ in range(self.params["channels"])],
            "data": [[] for _ in range(self.params["channels"])],
            "start_window": (raw_data[self.params["headers"]] & self.params["windmask"])
            >> self.last_bits,
        }

        self._parse_event_headers(curr_event, bitshifted_data)

        # Move through data one window at the time.
        num_headers = self.params["headers"]
        data_end = len(bitshifted_data) - self.num_footers
        step_size = self.num_wind_header + self.params["samples"]
        for window_idx in range(num_headers, data_end, step_size):

            # Extract the window header data
            channel = (raw_data[window_idx] & self.params["chanmask"]) >> self.params[
                "chanshift"
            ]
            window = (raw_data[window_idx] & self.params["windmask"]) >> self.last_bits
            curr_event["window_labels"][channel].append(window)

            curr_event["data"][channel].extend(
                bitshifted_data[
                    window_idx + 1 : window_idx + (self.params["samples"] + 1)
                ]
            )

        curr_event["data"] = np.array(curr_event["data"])

        return curr_event

    def _validate_input_package(self, in_data):
        """Validate the input data.

        Raises:
            BadDataException if:
            - there is no data.
            - there is not a en event amount of bytes, since 8-bits are combined to 16-bit words.
            - the data is not divisable by sample_num + window_headers.
        """
        if not isinstance(in_data, dict):
            raise TypeError("raw_data is not a dict it's a %s", type(in_data))

        if "rawdata" not in in_data or in_data["rawdata"] == b"":
            raise BadDataError("No Data Found")

        # Edge case, data doesn't contain valid words.
        if len(in_data["rawdata"]) % 2 != 0:
            raise BadDataError(
                f"Bad package no {in_data.get('pkg_num', 'unknown')}, "
                f"length is not an even amount of words, "
                f"len {len(in_data['rawdata'])}"
            )

        # Split this validation into it's own package.
        if self.params["model"] != "aardvarcv3":
            if (len(in_data["rawdata"]) // 2 - self.params["headers"] - 1) % (
                (self.params["samples"] + 1)
            ) != 0:
                raise BadDataError(
                    f"Bad package no {in_data.get('pkg_num', 'unknown')}"
                    f", length is not dividable by window sizes,"
                    f" len {len(in_data['rawdata'])}"
                )

    def _bitshift_data(self, raw_data):
        """Bitshift and bitmask the data.

        The 12-bit data can be positioned differently in the 16-bit data.
        By moving the last_bits and changing the data_bit_mask it's possbile to
        have data of various bit-length and position.
        """
        return (raw_data & self.data_bitmask) >> self.last_bits

    def _add_xaxis_to_event(self, event):
        """Adds an x-axis to the data.

        Based on the amount of channels and samples it will add a timeaxis
        per channel to the event dict.

        Returns:
            numpy array with sample numbers for each channel.
        """
        times = list()
        samples = self.samples
        channels = self.channels
        start_window = event["start_window"]
        for chan in range(0, channels):
            # LOGGER.debug("winds=%s", event["window_labels"][chan])
            winds = np.array(event["window_labels"][chan], dtype="int32")
            winds += self.windows * (winds < start_window)
            winds -= start_window

            tim = (samples * np.repeat(winds, samples)) + (
                np.ones((len(winds), samples), dtype=int)
                * np.arange(0, samples, dtype=int)
            ).flatten()

            times.append(tim)

        return times



from .default import TriggerController


class TriggerControllerTrbhm(TriggerController):
    """Trigger controller for TRBHM boards."""

    def __init__(self, board):
        super().__init__(board)

        self._num_chips = board.params.get("num_chips", 2)
        self._num_trig_chans_per_chip = board.params.get("num_trig_chans_per_chip", 4)

    def write_triggers(self):
        """Write the trigger values to the hardware.

        This writes the current set triggers to the hardware.
        By setting the trigger_offsets or the trigger_values
        this function is called automatically and the hardware is updated.
        """
        self._set_trigger_thresholds()
        self._set_wbiases()

        return True

    def _set_trigger_thresholds(self):
        """Sets the trigger values to those in the board object and writes to hardware."""
        trigger_values = self.board.trigger_values
        self._write_analog_registers_per_chip("trigger_threshold_{:02}", trigger_values)

    def _set_wbiases(self):
        """Set the wbias registers based on the board trigger values"""
        wbias = [self._wbias if x > 0 else 0 for x in self.board.trigger_values]
        self._write_analog_registers_per_chip("wbias_{:02}", wbias)

    def _write_analog_registers_per_chip(self, reg_name_format: str, values: list):
        """Write analog registers with a given register name format using the
        values provided. Writes values to both chips individually.

        Args:
            reg_name_format (str): register name format string, must take one integer argument.
            values (list): list of values to write. Should be all 8 values across both chips.
        """
        for chip in range(self._num_chips):
            self._select_chips([chip])

            for idx in range(self._num_trig_chans_per_chip):
                val = values[idx + chip * self._num_trig_chans_per_chip]
                self._write_analog_register(reg_name_format.format(idx), val)

        # TX needs to be restored after writing
        self._select_chips(list(range(self._num_chips)))

    def _select_chips(self, chips: "list[int]"):
        """Select which chips will be enabled for register writes."""
        self._write_control_register("ard_tx_en", sum([1 << x for x in chips]))



import logging
import time
from functools import partial

from naludaq.devices import LTC2990
from naludaq.helpers import type_name
from naludaq.helpers.exceptions import BadDataError

from .peripherals_controller import PeripheralsController

logger = logging.getLogger("naludaq.peripherals_hdsoc")


class AodsocPeripheralsController(PeripheralsController):
    def __init__(self, board):
        """AODSOC peripherals controller.

        Provides a way to read various I2C devices on the board.

        Args:
            board (Board): the board object.
        """
        super().__init__(board)
        self._reader_functions = {
            "current": self.read_current_of_chip,
            "vadjn0": partial(self.read_vadjn, 0),
            "vadjn1": partial(self.read_vadjn, 1),
            "chip_voltage": self.read_chip_voltage,
            "ldo_voltage": self.read_ldo_voltage,
        }

    def read_vadjn(self, chip: int) -> float:
        """Reads a VadjN measurement from the board.

        Args:
            chip (int): chip number, must be 0 or 1

        Returns:
            The measurement in Volts.
        """
        self._validate_chip_or_raise(chip)
        voltage = self._read_ltc2990_voltage(f"vadjn_{chip}")
        logger.debug("VadjN (%s): %s V", chip, voltage)
        return voltage

    def read_vadjp(self, chip: int = 0) -> float:
        """Get the vadjp value. It's not hooked up to the voltage monitor."""
        return 0

    def read_chip_voltage(self) -> float:
        voltage = self._read_ltc2990_voltage("chip_voltage")
        logger.debug("Chip voltage: %s V", voltage)
        return voltage

    def read_ldo_voltage(self) -> float:
        voltage = self._read_ltc2990_voltage("ldo_voltage")
        logger.debug("Chip voltage: %s V", voltage)
        return voltage

    def read_current_of_chip(self) -> float:
        v_ldo = self.read_ldo_voltage()
        v_chip = self.read_chip_voltage()
        return self._convert_voltage_to_current(v_ldo - v_chip)

    def _read_ltc2990_voltage(self, param_name: str, num_samples=1):
        """Reads a voltage from the LTC2990.

        Args:
            param_name (str): the name of the param in the board YAML.
            num_samples (int): number of sample points to average.
                Measurements jitter a bunch, so averaging is helpful.

        Returns:
            The voltage in Volts.

        Raises:
            BadDataError if the voltage cannot be read.
            NotImplementedError if the given parameter name is invalid.
        """
        try:
            params = self._get_peripheral_params(param_name)
            addr, chan = params["addr"], params["chan"]
        except (KeyError, NotImplementedError):
            raise NotImplementedError(
                f'Peripheral measurement "{param_name}" is not defined'
            )

        monitor = LTC2990(self._board, addr)

        samples = []
        for _ in range(num_samples):
            monitor.trigger_conversion()  # Important! Conversion needed before reading
            time.sleep(0.01)
            if chan == "vcc":
                try:
                    voltage = monitor.read_vcc()
                except BadDataError:
                    raise
            else:
                try:
                    voltage = monitor.read_voltage(chan)
                except BadDataError:
                    raise

            samples.append(voltage)

        return sum(samples) / len(samples)

    @staticmethod
    def _validate_chip_or_raise(chip: int):
        """Checks whether a given chip number is 0 or 1.

        Raises:
            TypeError: if chip is not an int
            ValueError: if chip is not valid
        """
        if not isinstance(chip, int):
            raise TypeError(f'Chip must be an int, not "{type_name(chip)}"')
        if chip not in [0, 1]:
            raise ValueError(f"Chip must be 0 or 1, not {chip}")



"""HDSoC is a special child in terms of how data is collected.
In order to generate pedestals, a sufficient amount of data needs
to be collected for every single absolute window in each channel.
However, the way data is returned with HDSoC
"""
import logging
import time
from collections import deque
from logging import getLogger

import numpy as np

from naludaq.controllers import get_board_controller, get_readout_controller
from naludaq.helpers.exceptions import OperationCanceledError, PedestalsDataCaptureError
from naludaq.tools.pedestals.pedestals_controller import PedestalsController
from naludaq.tools.waiter import EventWaiter

LOGGER = logging.getLogger(__name__)  # pylint: disable=invalid-name


from naludaq.controllers import get_board_controller, get_readout_controller
from naludaq.helpers.exceptions import NotAValidEvent, PedestalsDataCaptureError
from naludaq.helpers.helper_functions import event_transfer_time
from naludaq.tools.pedestals.pedestals_controller import PedestalsController

LOGGER = getLogger("naludaq.hdsoc_pedestals_generator")


class HDSoCPedestalsGenerator(PedestalsController):
    """HDSoC pedestals generator with firstwindow fix.

    This class is a fix for the HDSoC FW bug that causes the first window
    to be corrupted. This class will capture one extra window and discard.

    The HDSoC has a bug where you can only capture a certain amount of events before
    it needs a restart.

    This pedestals generator will work around that by starting/stopping readout multiple times per block.
    """

    def __init__(
        self,
        board,
        num_captures: int = 10,
        num_warmup_events: int = 10,
        channels: "list[int]" = None,
    ):
        super().__init__(board, num_captures, num_warmup_events, channels=channels)
        self.margin: float = 1.2
        wait_overhead = 1.0
        self._timeout: float = event_transfer_time(
            board,
            windows=board.params.get("pedestals_blocks", 16),
            channels=len(self._channels),
            margin=self.margin,
            overhead=wait_overhead,
        )
        self._attempts = 5
        self._first_win_fix = self.board.params.get("first_window_fix", True)

    def _capture_block(self, block, block_size, num_captures, num_warmup_events):
        """Capture data for a single block.

        This function overrides the parent class to capture data for a single block.

        This function works around the HDSoC FW bug to capture data
        it will capture as much data it can per run, then restart the
        readout and capture again.
        """
        num_captures = num_captures or self.num_captures
        num_warmup_events = num_warmup_events or self.num_warmup_events
        total_captures = num_captures + num_warmup_events
        block_size = block_size
        if self._first_win_fix:
            capture_size = block_size + 1  # +1 to capture the first window fix
        else:
            capture_size = block_size

        self.validated_data = deque(maxlen=num_captures)
        lb = self._get_lookback(block, block_size)
        rc = get_readout_controller(self.board)
        rc.set_read_window(
            windows=capture_size,
            lookback=lb,
        )

        self.block_buffer = []
        self._daq.output_buffer = deque(maxlen=total_captures)
        needed_amount = total_captures
        self._daq.start_capture()
        self._start_capture()
        expected_block = self._calculate_expected_block(block, block_size)
        while len(self.block_buffer) < needed_amount:
            try:
                next_event = self._next_validated_event_or_raise(
                    self._daq.output_buffer,
                    self._timeout,
                    self._attempts,
                    expected_block,
                )
            except PedestalsDataCaptureError:
                self._daq.stop_capture()
                self._daq.stop_workers()
                time.sleep(0.5)
                self._stop_capture()
                raise
            except OperationCanceledError:
                LOGGER.debug("Pedestals generation canceled")
                break

            self.block_buffer.append(next_event)

        if self._first_win_fix:
            self._remove_first_window(self.block_buffer)

        self._daq.stop_capture()
        self._daq.stop_workers()
        time.sleep(0.5)
        self._stop_capture()

        if self._store_warmup_events:
            self._warmup_data = self.block_buffer[:num_warmup_events]

        self.validated_data.extend(self.block_buffer[num_warmup_events:needed_amount])

        return True

    def _next_validated_event_or_raise(
        self, buffer, timeout: float, attempts: int, expected_block: list[int]
    ) -> dict:
        """Guaranteed to either return a valid event, or raise an error.
        This method must be called while capturing.

        Args:
            buffer (deque): daq output buffer
            timeout (float): expected timeout in seconds
            attempts (int): number of attempts
            expected_block (list[int]): list of expected windows in this block.

        Returns:
            dict: the event.

        Raises:
            PedestalsDataCaptureError: if a valid event could not be read.
            OperationCanceledError: if pedestals generation was canceled
        """
        event = None
        for i in range(attempts):
            self._raise_if_canceled()  # exit point for canceling
            self._toggle_trigger_or_raise()

            try:
                self._wait_for_event_or_raise(buffer, timeout)
                event = buffer.popleft()
            except (TimeoutError, IndexError):
                LOGGER.warning("Timed out while waiting for an event")
                continue

            if self._validate_event(event, expected_block):
                LOGGER.debug("Got a validated event")
                break
        else:
            msg = (
                "Failed to capture an event. The board may be unresponsive "
                "or need to be power cycled/reinitialized."
            )
            LOGGER.error("Failed to capture an event!")
            raise PedestalsDataCaptureError(msg)

        return event

    def _toggle_trigger_or_raise(self):
        """Toggle the external trigger or raise ``PedestalsDataCaptureError`` on failure"""
        try:
            get_board_controller(self.board).toggle_trigger()
        except Exception as e:
            msg = (
                "Pedestals data capture failed: software trigger could not "
                "be sent. Is the board connected?"
            )
            raise PedestalsDataCaptureError(msg) from e

    def _wait_for_event_or_raise(self, buffer, timeout):
        """Wait for an event or raise an error. If the function
        returns it is guaranteed that an event was read out.

        Args:
            buffer (Iterable): the daq output buffer
            timeout (float): timeout

        Raises:
            PedestalsDataCaptureError: if the event cannot be read in time.
        """
        waiter = EventWaiter(buffer, amount=len(buffer) + 1, timeout=timeout)
        try:
            waiter.start(blocking=True)
        except TimeoutError:
            raise
        except Exception as e:
            msg = "Pedestals data capture failed due to unknown reason"
            raise PedestalsDataCaptureError(msg) from e

    def _get_lookback(self, block, block_size):
        """Get lookback for the block."""
        if self._first_win_fix:
            output = block * (block_size) - 1
            if output < 0:
                output = self.board.params["windows"] - 1
        else:
            output = block * (block_size)
        return output

    def _validate_event(self, event, expected_block):
        """Returns true if the window labels matches the expected_block.

        The validation matches the received window_labels with the expected
        window labels to make sure the block contains only data from that
        block. The firmware buffer sometimes contains more data from a previous block
        this makes sure it doesn't enter the next block buffer.

        Args:
            event(dict): event to validate
            expected_block(list): a list of expected window numbers

        Returns:
            True if validated, False if the events windows doesn't match expected.

        Raises:
            NotAValidEvent if the event doesn't have a window_labels key.
        """
        try:
            for channel, window_label in enumerate(event["window_labels"]):
                if channel not in self._channels:
                    continue
                if not np.all(window_label[1:] == expected_block):  # Skip first window
                    LOGGER.debug(
                        "Expected: %s != returned: %s",
                        expected_block,
                        window_label[1:],
                    )
                    return False
        except Exception as error_msg:
            raise NotAValidEvent(f"Event validataion failed due to: {error_msg}")
        return True

    def _select_all_channels(self):
        """Pedestals must be generated over all channels."""
        channels_to_read = [x for x in range(self.board.params["channels"])]
        return channels_to_read

    def _start_capture(self):
        """Start board and daq capture"""
        readout_settings = {
            "trig": "ext",
            "lb": "forced",
            # "acq": "pedsub",
            # "dig_head": False,
            # "ped": "zero",
            # "readoutEn": True,
            # "singleEv": False,
        }
        get_board_controller(self.board).start_readout(**readout_settings)

    def _stop_capture(self):
        """Stop board and daq"""
        get_board_controller(self.board).stop_readout()

    def _remove_first_window(self, buffer):
        """Remove first window from buffer"""
        samples = self.board.params.get("samples", 32)
        for event in buffer:
            for li, lbl in [
                (i, lbl) for i, lbl in enumerate(event["window_labels"]) if len(lbl) > 0
            ]:
                event["window_labels"][li] = lbl[1:]
            for li, dat in [
                (i, dat) for i, dat in enumerate(event["data"]) if len(dat) > 0
            ]:
                event["data"][li] = dat[samples:]



from naludaq.backend.managers.base import Manager


class BoardIoManager(Manager):
    def __init__(self, board):
        """Utility for I/O with a board connected to the backend.

        Args:
            context (Context): context used to communicate with the backend.
        """
        super().__init__(board)

    def write(self, command: "str | bytes | list[str | bytes]"):
        """Writes a non-read command to the board.

        Args:
            command (str | bytes): command to send.
        """
        self.write_all([command])

    def write_all(self, commands: "list[str | bytes]"):
        """Writes a list of non-read commands to the board.

        Args:
            commands (list[str | bytes]): the commands to send.
        """
        if len(commands) == 0:
            raise ValueError("Need at least one command")
        commands = [c.hex() if isinstance(c, bytes) else c for c in commands]
        self.context.client.put(
            "/board/raw",
            json={"packages": commands},
        )

    def read(self, command: "str | bytes") -> bytes:
        """Sends a read command to the board and gets the response.

        Args:
            command (str | bytes): read command to send.

        Returns:
            bytes: the response.
        """
        return self.read_all([command])[0]

    def read_all(self, commands: "list[str | bytes]") -> list[bytes]:
        """Sends several read commands to the board and retrieves the responses.

        Args:
            commands (list[str | bytes]): commands to send

        Returns:
            list[bytes]: The responses from the board
        """
        if len(commands) == 0:
            raise ValueError("Need at least one command")
        commands = [c.hex() if isinstance(c, bytes) else c for c in commands]
        response = self.context.client.get_json(
            "/board/raw",
            json={"packages": commands},
        )

        return [bytes.fromhex(a) for a in response["packages"]]



"""IO module

Collection of IO functions to support disk IO.
"""
import csv
import gzip
import logging
import os
import pickle
import re

import numpy as np

from naludaq.helpers.exceptions import InvalidClockFileError
from naludaq.io.csv_writer import get_csv_writer

LOGGER = logging.getLogger(__name__)


class IOManager:
    """
    IO Controller handles all IO interactions with csv, yaml, etc. files for
    purposes such as exporting events and generating config files.

    Attributes:
        params (dict): Dictionary of parameters. Must include key: ``model``.
    """

    def __init__(self, params: dict = None):
        self.params = params
        self.model = None
        if params is not None:
            self.model = params.get("model", None)

    def export_acq_to_bin(self, filename: str, acqs: list):
        """Exporting multiple acquisitions to a bin file.

        Args:
            filename (str): full filepath inlcuding directory and filename.
            acqs (list): list of acquisitions to export data from.

        Output:
            bin file written to disk.
        """
        LOGGER.debug("Exporting %s acquisitions to bin", len(acqs))

        if not isinstance(acqs[0], dict):
            events = self._compress_acquisitions(acqs)
        else:
            events = acqs

        path, _ = os.path.split(filename)
        if not os.path.isdir(path):
            raise NotADirectoryError(f"Not a valid directory: {path}")

        with open(filename, mode="wb") as evt_file:
            pickle.dump(events, evt_file)

    def export_acq_to_csv(
        self, filename: str, acqs: list, max_channels=None, time_per_channel=False
    ):
        """Exporting multiple acquisitions to a csv file.
        Args:
            filename (str): full filepath inlcuding directory and filename.
            acqs (list): list of acquisitions to export data from.
            max_channels (int): set the maximum channels to readout.
            time_per_channel (bool): export time per channel rather than per row.

        Output:
            csv file written to disk.
        """
        csv_writer = get_csv_writer(self.params)
        csv_writer.export_acq_to_csv(filename, acqs, max_channels, time_per_channel)

    @staticmethod
    def _compress_acquisitions(acquisitions):
        """Converts multiple acquisitions into a single
        acquistion.

        Args:
            acquisitions (list): A list of acquisition

        Returns:
            acquisition (list): Combined acquisition of all events
        """
        events = list()
        for acq in acquisitions:
            events.extend(acq)

        return events

    @staticmethod
    def import_data_from_csv(
        filename: str, channels: list
    ):  # TODO(Marcus): Update for upac
        """
        Reads a csv file that has stored events and stored that data into a
        variant of the acquisition structure, called a kacquisition. (Name is a WIP)
        The difference is that a kacq flattens all events into a single list

        Args:
            filename (str): Filename
            channels (list): Specific channels to parse out of the csv

        """

        if not isinstance(channels, list):
            raise TypeError(f"Expected chans to be list, got {type(channels)}")

        with open(filename) as open_file:
            csv_f = csv.reader(open_file)

            acq = {
                "evt_num": [],
                "windnum": [],
                "data": [],
                "time": [],
            }

            for chan in channels:
                acq["evt_num"].append([])
                acq["windnum"].append([])
                acq["data"].append([])
                acq["time"].append([])

            _ = next(csv_f)
            if _[0] == "acq_num":
                for row in csv_f:
                    for index, chan in enumerate(channels):
                        acq["evt_num"][index].append(int(row[1]))
                        acq["windnum"][index].append(int(row[(chan * 3) + 2]))
                        acq["data"][index].append(int(float(row[(chan * 3) + 3])))
                        acq["time"][index].append(int(row[(chan * 3) + 4]))
            else:
                for row in csv_f:
                    for index, chan in enumerate(channels):
                        acq["evt_num"][index].append(int(row[0]))
                        acq["windnum"][index].append(int(row[(chan * 3) + 1]))
                        acq["data"][index].append(int(float(row[(chan * 3) + 2])))
                        acq["time"][index].append(int(row[(chan * 3) + 3]))

        return acq

    @staticmethod
    def import_data_from_bin(filename: str):
        """
        Reads a bin file that has stored events in a acquisition. Bin file has
        been pickled.

        Args:
            filename (str): Filename
            channels (list): Specific channels to parse out of the csv

        """

        with open(filename, mode="rb") as open_file:
            acq = pickle.load(open_file)

        return acq

    def import_data_from_directory(self, directory: str, chans: list):
        """Gets data from all csv files in a directory.

        Primarily for importing frequency sweep results.

        Args:
            directory (str): Path to folder where files to be imported are located
            chans (list): List of channels to import.

        Returns:
            freq_manifold (list): List of frequecies labeled in the filenames
            data_manifold (list): List of events that was imported from the csv file

        Raises:
            TypeError if chans is not list
        """

        data_manifold = []
        freq_manifold = []
        for file in os.listdir(directory):
            filename = os.fsdecode(file)
            if filename.endswith(".csv"):
                data_manifold.append(
                    self.import_data_from_csv(
                        os.path.join(directory, filename),
                        channels=chans,
                    )
                )
                freq_manifold.append(int(file[:-4]))

        return freq_manifold, data_manifold

    def export_sweep_results(self, directory: str, sweep_results: list):
        """Takes the results of a parameter sweep and stores the results as a collection of csvs
        inside of a folder.

        Args:
            directory (str): Location where to store all the csvs
            sweep_results (list): List of data acquisitions from varying parameters
        """

        for item in sweep_results:
            freq_num = str(int(item[0] / 1000000))
            while len(freq_num) < 4:
                freq_num = "0{0}".format(freq_num)

            filename = directory + "/" + freq_num + ".csv"
            self.export_acq_to_csv(filename, [item[1]])

    def export_pedestals_csv(self, board, filename: str, pedestals: dict):
        """Export all data from the pedestals to CSV.

        Please note this will only export the data, not the information regarding the pedestals.
        This function also assumes filename is in the correct format.
        It will override any existing file.

        Args:
            board (Board): A board object with the pedestals data attached to it.
            filename (path): Filename for the csv file.
        """
        get_csv_writer(self.params).export_pedestals_csv(board, filename, pedestals)


def save_calibration_data(board, filename):
    """Save calibration data from the board object.

    Args:
        board: good ol' Board object
        filename: filepath and filename
    """
    caldata = getattr(board, "caldata", None)

    if not isinstance(caldata, dict):
        raise TypeError(f"pedestals must be a dict, got {type(caldata)}")
    if filename is None:
        raise TypeError("Supplied pathname is NoneType.")
    path, _ = os.path.split(filename)
    if not os.path.isdir(path):
        raise NotADirectoryError(f"Not a valid directory: {path}")

    try:
        sfile = gzip.GzipFile(filename, "w", compresslevel=4)
        pickle.dump(caldata, sfile, protocol=pickle.HIGHEST_PROTOCOL)
    except IOError as error_msg:
        LOGGER.error("File could not be created. %s", error_msg)
    except pickle.PicklingError as error_msg:
        LOGGER.error("Saving pedestals failed: %s", error_msg)


def load_calibration_data(filename):
    """Load calibration data.

    Args:
        filename: the path to the file containing calibration data

    Returns:
        caldata dictionary.
    """
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"No file found: {filename}")

    try:
        sfile = gzip.GzipFile(filename, "r")
        outped = pickle.load(sfile)
    except EOFError as error_msg:
        LOGGER.error("Reached the end of the file, msg: %s", error_msg)
    except IOError as error_msg:
        LOGGER.error("File could not be loaded. %s", error_msg)
    except pickle.UnpicklingError as error_msg:
        LOGGER.error("Loading pedestals failed: %s", error_msg)
    else:
        if not isinstance(outped, dict):
            raise TypeError(f"Not a valid Calibration file: {filename}")

        return outped


def save_timingcal_data(board, filename):
    """Save timing calibration data from the board object.

    Args:
        board: good ol' Board object
        filename: filepath and filename

    Raises:
        TypeError if
        NotADirectoryError if the filename is not in a valid path
    """
    caldata = getattr(board, "timingcal", None)
    if caldata is None:
        raise TypeError(f"Timingcal data is not valid, got {type(caldata)}")

    if filename is None:
        raise TypeError("Supplied pathname is NoneType.")
    path, _ = os.path.split(filename)
    if not os.path.isdir(path):
        raise NotADirectoryError(f"Not a valid directory: {path}")

    try:
        sfile = gzip.GzipFile(filename, "w", compresslevel=4)
        pickle.dump(caldata, sfile, protocol=pickle.HIGHEST_PROTOCOL)
    except IOError as error_msg:
        LOGGER.error("File could not be created. %s", error_msg)
    except pickle.PicklingError as error_msg:
        LOGGER.error("Saving pedestals failed: %s", error_msg)


def load_timingcal_data(filename):
    """Load timing calibration data.

    Args:
        filename: the path to the file containing timing calibration data

    Returns:
        timingcal data

    Raises:
        FileNotFoundError if filename is not a file.
    """
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"No file found: {filename}")

    try:
        sfile = gzip.GzipFile(filename, "r")
        output = pickle.load(sfile)
    except EOFError as error_msg:
        LOGGER.error("Reached the end of the file, msg: %s", error_msg)
    except IOError as error_msg:
        LOGGER.error("File could not be loaded. %s", error_msg)
    except pickle.UnpicklingError as error_msg:
        LOGGER.error("Loading pedestals failed: %s", error_msg)
    else:
        return output


def load_clockfile(filename):
    """Turns a .txt file into a list, used for clock file loading.

    Returns:
        list of the loaded clock commands.

    Raises:
        FileNotFoundError if filename is not a valid text file.
        Error ifr loadtxt function fails for an unknown reason (The numpy function is fragile).
        TypeError if clock data type is not a str
        InvalidClockFileError if the clock data format is incorrect
    """
    if not os.path.exists(filename) or not os.path.split(filename)[-1].endswith(".txt"):
        raise FileNotFoundError(
            f"File {filename} does not exist. Please provide the path to a valid clockfile."
        )

    try:
        data = np.loadtxt(filename, dtype=str, delimiter=",")
    except Exception:
        raise
    validate_clock_data_or_raise(data)
    return data


def validate_clock_data_or_raise(clock_data):
    header = ["Address", "Data"]
    try:
        if not np.array_equal(clock_data[0], header):
            raise
    except:
        raise InvalidClockFileError(
            "Invalid clock data header, expected: 'Address', 'Data'"
        )
    for line, value in enumerate(clock_data[1:]):
        try:
            val_addr = re.fullmatch("0x[a-fA-F0-9]{4}", value[0])
            val_data = re.fullmatch("0x[a-fA-F0-9]{2}", value[1])
        except TypeError:
            raise TypeError(
                f"Clock data on line {line + 1} is of the wrong type, got address type {type(value[0])} and data type {type(value[1])}"
            )
        except (IndexError, KeyError):
            raise InvalidClockFileError(
                f"Clock data on line {line + 1} is of incorrect length or type"
            )
        if not val_addr or not val_data:
            raise InvalidClockFileError(
                f"Invalid clock data on line {line+1}: addr={value[0]}, data={value[1]}"
            )



from typing import Sequence

import numpy as np

from naludaq.helpers.exceptions import InvalidParameterFile

from .i2c_device import I2CDevice


class EEPROM(I2CDevice):
    def __init__(self, board):
        """Class for controlling the M24M01 EEPROM.

        The M24M01 is a 1-megabit (128 kilobyte) I2C bus EEPROM. The full address space
        is accessible to the user.

        A portion of memory is addressed using a 17 bit address. The lower 16 bits
        are included in read/write commands, while the highest bit is actually the
        lowest bit of the device address.

        Args:
            board (Board): the board object.

        Raises:
            InvalidParameterFile: if the YAML does not contain a definition for the EEPROM
        """
        self._validate_eeprom_params_or_raise(board.params)
        self._eeprom_params = board.params["peripherals"]["eeprom"]
        self._default_addr = self._eeprom_params["addr"]
        super().__init__(board, self._default_addr)

        self._default_addr = self._eeprom_params["addr"]
        self._capacity = self._eeprom_params["capacity"]
        self._segments = self._eeprom_params["segments"]
        self._page_size = self._eeprom_params["page_size"]

        self._max_io_bytes = (
            4  # max bytes that can be sent/received with one I2C command
        )
        self._byte_order = ">"

    def store_array(self, addr: int, arr: "Sequence | np.ndarray", dtype: np.dtype):
        """Store an array into the EEPROM

        Args:
            addr (int): the starting address of the array
            arr (Sequence | np.ndarray): the array to store. For safety, the dtype
                of this array is ignored and the array is converted to the dtyep given
                instead.
            dtype (np.dtype): the dtype to store the array as. An object dtype cannot
                be used.

        Raises:
            TypeError: if the given combination of array and dtype results in an
                array that cannot be stored.
            I2CError: if there was a problem communicating with the device
        """
        dtype = self._convert_type_to_dtype_or_raise(dtype)
        arr = self._convert_array_dtype_or_raise(arr, dtype).flatten()
        self._validate_address_block_or_raise(addr, arr.size * dtype.itemsize)

        self.store_raw_data(addr, arr.tobytes())

    def load_array(self, addr: int, num_elements: int, dtype: np.dtype) -> np.ndarray:
        """Load an array from the EEPROM.

        The size of the block read depends on the number of elements in the
        array and the element data type.

        Args:
            addr (int): the starting address of the array
            num_elements (int): the number of elements in the array
            dtype (np.dtype): the data type for elements in the array

        Returns:
            np.ndarray: a 1-D array with the given data type.

        Raises:
            I2CError: if there was a problem communicating with the device
        """
        if not isinstance(num_elements, int):
            raise TypeError("Number of elements must be an int")
        dtype = self._convert_type_to_dtype_or_raise(dtype)
        self._validate_address_block_or_raise(addr, num_elements * dtype.itemsize)

        data = self.load_raw_data(addr, num_elements * dtype.itemsize)
        return np.frombuffer(data, dtype)

    def store_raw_data(self, addr: int, data: bytes):
        """Write raw data to the EEPROM memory.

        Args:
            addr (int): pointer to the start of the block to write.
            data (bytes-like): data to store in the EEPROM. Can be any size, as long
                as the data fits into the location provided (see YAML).

        Raises:
            I2CError: if there was a problem communicating with the device
        """
        if not isinstance(data, (bytes, bytearray)):
            raise TypeError("Data address must be bytes-like")
        self._validate_address_block_or_raise(addr, len(data))

        for start_addr, data_chunk in self._iter_data_chunks(addr, data):
            addr_bytes = self._get_data_addr_bytes(start_addr)
            self._set_addr_from_byte_loc(start_addr)
            self.send_write_command(addr_bytes + data_chunk)

    def load_raw_data(self, addr: int, amount: int) -> bytes:
        """Reads data from the EEPROM.

        Args:
            addr (int): pointer to the start of the block to read.
            amount (int): length of data to read. Can be any length, as long as the
                block does not exceed the boundaries of the storage array (see YAML).

        Returns:
            bytes: the data read from the EEPROM.

        Raises:
            I2CError: if there was a problem communicating with the device
        """
        self._validate_address_block_or_raise(addr, amount)

        result = bytearray()
        chunk_size = self._max_io_bytes
        for start_addr in range(addr, addr + amount, chunk_size):
            self._set_addr_from_byte_loc(start_addr)

            # NOTE: ack-checking after writing causes too much delay which ends up trashing the response
            self.send_write_command(
                self._get_data_addr_bytes(start_addr), check_ack=False
            )
            self.send_read_command(bytes([0xFF] * chunk_size), check_ack=False)
            response = self._read_response_registers()

            response = self._discard_ack_bits(response)[1 : chunk_size + 1]
            result += response

        return bytes(result[:amount])

    def _get_data_addr_bytes(self, data_addr: int):
        """Get the address for data as a `bytes` object.

        Args:
            data_addr (int): address of the data being written

        Returns:
            bytes: the 2-byte address
        """
        return data_addr.to_bytes(2, "big")

    def _set_addr_from_byte_loc(self, byte_loc: int):
        """Sets the device address based on the byte location. The lowest bit of
        the device address is the highest bit of the data address.
        """
        self._addr = self._default_addr + ((byte_loc >> 16) & 1)

    def _validate_address_block_or_raise(self, start: int, length: int):
        """Makes sure a block of addresses is fully contained within the range
        of valid addresses.

        Args:
            start (int): start address
            length (int): block length

        Raises:
            TypeError: if the start address or length is not an int.
            ValueError: if any portion of the block is out of bounds.
        """
        if not isinstance(start, int):
            raise TypeError("Data address must be int")
        if not isinstance(length, int):
            raise TypeError("Data length must be int")
        if length <= 0:
            raise ValueError("Data length cannot be negative")
        if not length <= start + length <= self._capacity:
            raise ValueError(f"Address block is out of bounds")

    @staticmethod
    def _validate_eeprom_params_or_raise(params: dict) -> int:
        """Validate EEPROM params in YAML.

        Args:
            params (dict): board params

        Raises:
            InvalidParameterFile: if the params are invalid.
        """
        eeprom_params = params.get("peripherals", {}).get("eeprom", None)
        if eeprom_params is None:
            raise InvalidParameterFile(f"Board does not have EEPROM support")

        required_entries = {
            "addr": int,
            "capacity": int,
            "segments": dict,
        }
        for name, type_ in required_entries.items():
            value = eeprom_params.get(name, None)
            if value is None:
                raise InvalidParameterFile(f'EEPROM params missing key "{name}"')
            if not isinstance(value, type_):
                raise InvalidParameterFile(
                    f'EEPROM params value for "{name}" has wrong type {type(value).__name__}, expected {type.__name__}'
                )

    def _convert_type_to_dtype_or_raise(self, dtype) -> np.dtype:
        try:
            dtype = np.dtype(dtype).newbyteorder(self._byte_order)
        except TypeError:
            raise
        if dtype == object:
            raise TypeError('Data type "object" cannot be safely stored')
        return dtype

    @staticmethod
    def _convert_array_dtype_or_raise(arr, dtype: np.dtype):
        arr = np.array(arr)
        try:
            arr = np.array(arr).astype(dtype)
        except ValueError as e:
            raise ValueError("Invalid array and/or datatype") from e
        if arr.dtype == object:
            raise TypeError('Data type "object" cannot be safely stored')
        return arr

    def _iter_data_chunks(self, start_addr: int, data: bytes):
        """Iterate over data in chunks. Yields (starting address, chunk data) for
        each chunk.

        The device doesn't allowing write operations that cross page boundaries,
        so this function splits a chunk that does cross boundaries into two smaller
        chunks that live on either side of the page boundary.

        If the data is not evenly divisible by the chunk size, then the last chunk
        is made smaller.
        """
        page_size = self._page_size
        chunk_size = self._max_io_bytes

        for i in range(0, len(data), chunk_size):
            # page numbers determined by edges of data addresses
            upper_page = (start_addr + i + chunk_size - 1) // page_size
            lower_page = (start_addr + i) // page_size

            if upper_page != lower_page:
                addr_boundary = upper_page * page_size
                data_boundary = addr_boundary - start_addr - i
                yield start_addr + i, data[i:data_boundary]
                yield addr_boundary, data[data_boundary : i + chunk_size]
            else:
                yield start_addr + i, data[i : i + chunk_size]



from naludaq.daq.workers.packager.worker_packager import PackagerLight
from naludaq.daq.workers.packager.worker_packager_debug import DebugPackager
from naludaq.daq.workers.packager.worker_packager_hdsoc import HDSoCPackager


def get_packager(board, *args, **kwargs):
    """Returns a Packager object for the given board."""
    if board.params["model"] in ["hdsocv1", "hdsocv1_evalr1", "hdsocv1_evalr2"]:
        return HDSoCPackager(board, *args, **kwargs)
    else:
        return PackagerLight(*args, **kwargs)
    # else:
    #     raise ValueError('Unknown board model: {}'.format(board.params['model']))



"""Controlls parameters of the communication.

Collection of tools to change the paramters of the communication.
Can be stateless.
"""
import time
from logging import getLogger

from naludaq.backend.managers.connection import ConnectionManager
from naludaq.backend.managers.io import BoardIoManager
from naludaq.backend.models.device import DeviceType
from naludaq.communication import ControlRegisters
from naludaq.controllers.board import BoardController
from naludaq.controllers.controller import Controller

LOGGER = getLogger(__name__)


class ConnectionController(Controller):
    """Controls connection settings.

    Used to alter the connection in any way.
    Change speed, ports, reset connection to a known start.

    TODO: Separate UART only functionality.

    """

    def change_speed(self, sync_int, rate, divider, sync_addr=0x00):
        """
        Attempts to change the baudrate!

        Divider sets it so: (25MHz clock / 16clks/bit)/(divider+1) = baudrate

        NOTE: It is a 100MHz clock now, so the math changes accordingly
            Also it looks like the baud gets quantized on the CPU side at higher rates, so keep that in mind

        Command is DDDDVVVV where V is the divider value (max 256)

        Args:
            sync_int (int): expected integer return when syncing with the board.
            rate (int): valid baudrate
            divider (int): baud divider
        """
        if self.board.using_new_backend:
            return self._change_speed_new(rate, divider)
        if not self.board.connection.is_uart():
            raise ConnectionError
        divider_val = f"{divider:04X}"
        cmd = "DDDD" + divider_val
        self.board.connection.send(cmd)
        time.sleep(
            0.1
        )  # It's essential to let the board update the baudrate, or it'll get stuck.
        self.board.connection.baud = rate

        # Resyncing the connection.
        for _ in range(3):
            if self.board.connection.resync(sync_int, sync_addr):
                LOGGER.debug(
                    "Baudrate synced successfully, divider: %s, baud: %s", divider, rate
                )
                break
        else:
            LOGGER.debug("Sync failed.")

    def set_speed_or_revert_to_default(self, speed=None):
        if self.board.using_new_backend:
            return self._set_speed_or_revert_to_default_new(speed)
        # Since the baud has been reset to default, set the baud to user settings.
        LOGGER.debug("SET SPEED: %s", self.board.connection_info["speed"])
        if not speed:
            speed = self.board.connection_info["speed"]

        if not self._validate_baud(speed):
            LOGGER.warning("Baudrate %s wasn't validated.", speed)
            baudparam = self.board.params["default_baud"]
        else:
            baudparam = [
                (name, divider)
                for name, divider in self.board.params["possible_bauds"].items()
                if int(name) == int(speed)
            ][0]

        LOGGER.debug("Setting baudrate to %s", baudparam)

        sync_int = self._get_sync_int()

        try:
            self.change_speed(sync_int, baudparam[0], baudparam[1])
            if self.board.model in ["upac32", "upaci", "zdigitizer"]:
                ControlRegisters(self.board)._update_software_register(
                    "baud_rate_divisor", baudparam[1]
                )
        except Exception as error_msg:
            raise ConnectionError(error_msg)

    def _validate_baud(self, baud: int):
        """Check if supplied baud is in boardparams.

        Args:
            baud(int): Baudrate

        Returns:
            (list) True if baudrate is valid for the selected board.
        """
        LOGGER.debug(
            "Validating using %s",
            [
                x == int(baud)
                for x, divider in self.board.params["possible_bauds"].items()
            ],
        )
        return any(
            [
                x == int(baud)
                for x, divider in self.board.params["possible_bauds"].items()
            ]
        )

    def reset_connection(self):
        """Matches the board baudrate and resets it to the default.

        This is useful if the application crashes or is in anyway
        not stopping as intended and the baudrate is not reset.
        This way it tries all baudrates available to the board
        if it works it resets the baudrate back to default.

        Returns:
            True if baudrate is reset to default.
        """
        if self.board.using_new_backend:
            return self._reset_connection_new()
        matched = False
        try:
            matched = self._match_baudrate()
        except Exception as error_msg:
            LOGGER.exception("Couldn't match the baudrate! Because: %s", error_msg)

        if matched:
            try:
                self.reset_speed()
            except ConnectionError as error_msg:
                LOGGER.exception("Baudrate not matched due to: %s", error_msg)
                return False
            else:
                return True
        return False

    def _get_sync_int(self):
        return self.board.registers["control_registers"]["identifier"]["value"]

    def _get_sync_addr(self):
        return self.board.registers["control_registers"]["identifier"]["address"]

    def _match_baudrate(self) -> bool:
        """Matches baudrate if the board is set different than program.

        IF the baord i set at a higher baudrate the program can't communicate.
        The only way around it is to guess and then try to lower the baudrate .
        Once the rate is lowered the init script can run and the speed can be increased.

        Returns:
            True if the baudrate is matched else False.
        """
        if self.board.using_new_backend:
            return self._match_baudrate_new()

        board = self.board
        LOGGER.info("Matching baudrate")
        possible_bauds = board.params["possible_bauds"]
        print(f"Trying {[x for x in possible_bauds.keys()]}")
        sync_int = self._get_sync_int()  # TODO: replace the speed selection system
        sync_addr = self._get_sync_addr()

        # early stop
        identifier = self._get_sync_response()
        if identifier == sync_int:
            # if self.board.connection.readReg(0) == sync_int:
            LOGGER.info("Baudrate already matched")
            return True

        LOGGER.debug(
            f"Identifier mismatch, hardware returned: {identifier} - expected: {sync_int}"
        )

        for _ in range(3):
            for baudrate, divider in possible_bauds.items():
                self.change_speed(sync_int, baudrate, divider, sync_addr)
                if board.model not in ["upac32", "upaci", "zdigitizer"]:
                    BoardController(board).reset_board()

                # Clear buffer
                time.sleep(0.05)
                identifier = self._get_sync_response()
                print(f"ident: {identifier} - {sync_int}")
                if identifier == sync_int:
                    # if self.board.connection.readReg(0) == sync_int:
                    LOGGER.info("Rate: %s locked", baudrate)
                    return True
                LOGGER.info("Rate: %s didn't lock, updating settings.", baudrate)
                time.sleep(0.1)
            time.sleep(1)
        return False

    def _get_sync_response(self):
        """Return a pre-defined identifier from the hardware."""
        self.configure_connection()
        identifier = -1
        try:
            identifier = ControlRegisters(self.board).read("identifier")["value"]
        except Exception as error_msg:
            LOGGER.debug("sync response failed due to: %s", error_msg)

        return identifier

    def reset_speed(self):
        """Set the baudrate to the default of the board.

        The default baudrate is stored with the default parameters for the board.
        The conversion between the divider and the baudrate is:
        int((clockrate / 16) / (baudrate) -1)

        Raises:
            ConnectionError if baudrate can't be set.
        """
        if self.board.using_new_backend:
            return self._reset_speed_new()
        board = self.board

        default_baud, divider = [
            (key, val) for key, val in board.params["default_baud"].items()
        ][0]
        sync_int = self._get_sync_int()

        try:
            self.change_speed(sync_int, default_baud, divider)
        except Exception as error_msg:
            raise ConnectionError(
                "Can't change baudrate on the selected connection: %s", error_msg
            )

    def verify_connection(self) -> bool:
        """Verify the connection to the hardware returns a valid response.

        Checks the first register on the board and checks if the response matches the expected.
        If it doesn't respond with the expected value the connection is likely not properly setup
        and/or there is garbage in the buffer.

        Returns:
            True if the response matches the expected response.
            False if any other value is returned.
        """
        return self.is_synced

    def reset_baud_rate(self, attempts: int = 2, num_null_commands: int = 20):
        """Sends a series of zeros ("null" commands) to reset the baud rate
        of the board

        This will only work for boards running firmware which supports this
        method of resetting the baud rate. Make sure a UART connection is open
        before running this method.

        Args:
            attempts (int): The number of times to attempt resetting
                the baud rate
            num_null_commands (int): The number of null/invalid commands to send

        Returns:
            True if the baud rate was successfully reset, or false otherwise
        """
        if self.board.using_new_backend:
            return self._reset_baud_rate_new(attempts, num_null_commands)
        try:
            ser = self.board.connection.ser
        except:
            raise ConnectionError(
                "Cannot reset the baud rate without a UART connection"
            )

        if not self.board.connection.is_open:
            raise ConnectionError(
                "Cannot reset the baud rate without a UART connection"
            )

        # Set the serial baud rate to the default value
        ser.baudrate = int(self.board.params["default_baudrate"])
        sync_int = self._get_sync_int()

        for i in range(attempts):
            LOGGER.debug("Trying to reset baud rate, attempt %d/%d", i + 1, attempts)

            # Send a bunch of zeros to reset the board's baud rate
            self.board.connection.send("00000000" * num_null_commands)
            time.sleep(0.05)

            # Check if the board is responsive
            identifier = self._get_sync_response()
            if identifier == sync_int:
                LOGGER.debug("Successfully reset baud rate")
                return True

        LOGGER.debug("Failed to reset baud rate")
        return False

    # =======================================================================
    # new stuff
    # =======================================================================
    @property
    def expected_sync_response(self) -> int:
        return ControlRegisters(self.board).registers["identifier"]["value"]

    @property
    def is_synced(self) -> bool:
        try:
            return self._get_sync_response() == self.expected_sync_response
        except:
            return False

    def configure_connection(self):
        """Set the appropriate registers so we can communicate with the
        board using the current connection.

        - For ethernet, this means setting tx_mode to ETH and updating the receiver address registers.
        - For non-ethernet, this means setting tx_mode to UART.
        """
        if (
            self.board.using_new_backend
            and ConnectionManager(self.board).device.type == DeviceType.UDP
        ):
            self._configure_ethernet()
        else:
            self._configure_non_ethernet()

    def _configure_non_ethernet(self):
        """Configure non-ethernet based (UART/FTDI/USB3) connections."""
        try:
            # enable UART, disable UDP
            self._write_control("tx_mode", 0)
            time.sleep(0.01)
        except:
            LOGGER.debug("Board does not have ethernet capabilities")

    def _configure_ethernet(self):
        """Configure an ethernet-based connection"""
        recv_ip, recv_port = self.board.connection_info["receiver_addr"]
        recv_octets = self._get_octets_or_raise(recv_ip)

        # enable UDP, disable UART
        self._write_control("tx_mode", 1)

        # set destination address
        self._write_control(
            "eth_dest_addr31_16", (recv_octets[0] << 8) | recv_octets[1]
        )
        self._write_control("eth_dest_addr15_0", (recv_octets[2] << 8) | recv_octets[3])
        self._write_control("eth_dest_port", recv_port)

        # select user destination address
        self._write_control("eth_addr_sel", 0b01)
        self._write_control("eth_port_sel", 0b01)

    def _change_speed_new(self, rate: int, divider: int):
        """New version of the same function"""
        self._validate_connection_or_raise()

        # It's essential to let the board update the baudrate, or it'll get stuck.
        BoardIoManager(self.board).write(f"DDDD{divider:04X}")
        time.sleep(0.1)
        ConnectionManager(self.board).device.baud_rate = rate

        # Resyncing the connection.
        for _ in range(3):
            if self._resync_new():
                LOGGER.debug(
                    "Baudrate synced successfully, divider: %s, baud: %s", divider, rate
                )
                break
        else:
            LOGGER.debug("Sync failed.")

    def _resync_new(self):
        """New version of the same function"""
        self._validate_connection_or_raise()
        LOGGER.debug("Syncing board...")
        ConnectionManager(self.board).device.clear_buffers()

        for i in range(0, 8):
            if self.is_synced:
                LOGGER.debug("synced, was off %s chars", i)
                return True

            LOGGER.debug("resync(): sending single character")
            BoardIoManager(self.board).write("FF")
            time.sleep(0.1)

        LOGGER.debug("Couldn't sync")
        return False

    def _set_speed_or_revert_to_default_new(self, speed: int = None):
        """New version of the same function"""
        self._validate_connection_or_raise()

        # Since the baud has been reset to default, set the baud to user settings.
        if not speed:
            speed = self.board.connection_info["baud_rate"]
        LOGGER.debug("SET SPEED: %s", speed)

        if not self._validate_baud(speed):
            LOGGER.warning("Baudrate %s wasn't validated.", speed)
            baudparam = self.board.params["default_baud"]
        else:
            divisor = {
                int(b): int(d) for b, d in self.board.params["possible_bauds"].items()
            }[int(speed)]
            baudparam = (int(speed), divisor)

        LOGGER.debug("Setting baudrate to %s", baudparam)

        try:
            self._change_speed_new(baudparam[0], baudparam[1])
        except Exception as error_msg:
            raise ConnectionError(error_msg)

    def _reset_connection_new(self):
        """Matches the board baudrate and resets it to the default.

        This is useful if the application crashes or is in anyway
        not stopping as intended and the baudrate is not reset.
        This way it tries all baudrates available to the board
        if it works it resets the baudrate back to default.

        Returns:
            True if baudrate is reset to default.
        """
        self._validate_connection_or_raise()
        matched = False
        try:
            matched = self._match_baudrate_new()
        except Exception as error_msg:
            LOGGER.exception("Couldn't match the baudrate! Because: %s", error_msg)

        if matched:
            try:
                self._reset_speed_new()
            except ConnectionError as error_msg:
                LOGGER.exception("Baudrate not matched due to: %s", error_msg)
                return False
            else:
                return True
        return False

    def _match_baudrate_new(self) -> bool:
        """Matches baudrate if the board is set different than program.

        IF the baord i set at a higher baudrate the program can't communicate.
        The only way around it is to guess and then try to lower the baudrate .
        Once the rate is lowered the init script can run and the speed can be increased.

        Returns:
            True if the baudrate is matched else False.
        """
        board = self.board
        LOGGER.info("Matching baudrate")
        possible_bauds = board.params["possible_bauds"]
        LOGGER.info(f"Trying {[x for x in possible_bauds.keys()]}")

        # early stop
        sync_response = self._get_sync_response()
        if sync_response == self.expected_sync_response:
            LOGGER.info("Baudrate already matched")
            return True

        LOGGER.debug(
            f"Identifier mismatch, hardware returned: {sync_response} - expected: {self.expected_sync_response}"
        )

        for _ in range(3):
            for baudrate, divider in possible_bauds.items():
                self._change_speed_new(baudrate, divider)
                BoardController(board).reset_board()

                # Clear buffer
                time.sleep(0.05)
                ConnectionManager(board).device.clear_buffers()

                sync_response = self._get_sync_response()
                LOGGER.info(
                    f"Sync response: 0x{sync_response:04X}, need: 0x{self.expected_sync_response:04X}"
                )
                if sync_response == self.expected_sync_response:
                    LOGGER.info("Rate: %s locked", baudrate)
                    return True
                LOGGER.info("Rate: %s didn't lock, updating settings.", baudrate)
                time.sleep(0.1)
            time.sleep(1)
        return False

    def _reset_speed_new(self):
        """New version of the same function"""
        self._validate_connection_or_raise()
        board = self.board

        default_baud, divider = [
            (key, val) for key, val in board.params["default_baud"].items()
        ][0]

        try:
            self._change_speed_new(default_baud, divider)
        except Exception as error_msg:
            raise ConnectionError(
                "Can't change baudrate on the selected connection: %s", error_msg
            )

    def _reset_baud_rate_new(self, attempts: int, num_null_commands: int):
        """New version of the same function"""
        self._validate_connection_or_raise()

        # Set the serial baud rate to the default value
        baud_rate = int(self.board.params["default_baudrate"])
        ConnectionManager(self.board).device.baud_rate = baud_rate

        for i in range(attempts):
            LOGGER.debug("Trying to reset baud rate, attempt %d/%d", i + 1, attempts)

            # Send a bunch of zeros to reset the board's baud rate
            BoardIoManager(self.board).write("00000000" * num_null_commands)
            time.sleep(0.05)

            # Check if the board is responsive
            if self.is_synced:
                LOGGER.debug("Successfully reset baud rate")
                return True

        LOGGER.debug("Failed to reset baud rate")
        return False

    def _validate_connection_or_raise(self):
        if not self.board.is_connected:
            raise ConnectionError("A connection is required")
        if (
            self.board.using_new_backend
            and not ConnectionManager(self.board).is_uart_based
        ):
            raise ConnectionError("Invalid connection type")
        if self.board.connection is not None and not self.board.connection.is_uart():
            raise ConnectionError("Invalid connection type")

    def _write_control(self, name: str, value: int):
        """Write a control register."""
        try:
            ControlRegisters(self.board).write(name, value)
        except ValueError as e:
            LOGGER.error("Failed to update control register due to: %s", e)

    def _get_octets_or_raise(self, ip: str) -> list[int]:
        """Try to parse the octets from an IPv4 address or raise an error.

        Args:
            ip (str): IPv4 address

        Returns:
            list[int]: list of ordered integer octets
        """
        if not isinstance(ip, str):
            raise TypeError("IP must be a string")
        try:
            octets = [int(o) for o in ip.split(".")]
        except:
            raise ValueError("Invalid IP address")
        if len(octets) != 4 or any(not 0 <= o <= 255 for o in octets):
            raise ValueError("Invalid IP address")
        return octets



"""ASoCv3Parser

Use parser to parse the data from the Hardware.
The boards sends a raw bitstream and it needs to be parsed into events.
The parser class is a tool to parse the raw bitstream of 8-bit chunks
into 16 bit words then extracting the 12-bit data, the header info,
channels and window number data.

This data is then returned in a dictionary.
The raw data is preserved.

This module was previously located under the naludaq.daq.workers
subpackage, and was relocated to improve the organization of the
naludaq package.
"""

import logging
import operator
from collections import defaultdict
from typing import List

import numpy as np

from naludaq.helpers.exceptions import BadDataError
from naludaq.parsers.parser import Parser

LOGGER = logging.getLogger("naludaq.hdsoc_parser")


class HDSoCParser(Parser):
    def __init__(self, params):
        super().__init__(params)
        self._stop_word = params.get("stop_word", b"\xfa\x5a")
        if isinstance(self._stop_word, str):
            self._stop_word = bytes.fromhex(self._stop_word)
        self._chan_mask = params.get("chanmask", 0x3F)
        self._chan_shift = params.get("chanshift", 0)
        self._abs_wind_mask = params.get("abs_wind_mask", 0x3F)
        self._evt_wind_mask = params.get("evt_wind_mask", 0x3F)
        self._evt_wind_shift = params.get("evt_wind_shift", 6)
        self._timing_mask = params.get("timing_mask", 0xFFF)
        self._timing_shift = params.get("timing_mask", 12)
        self._packet_size = 72

    def _validate_input_package(self, in_data):
        """HDSoC: Splits the input package into a list of packets,
        and validates each packet based on a fixed size. Returns
        a list of only valid packets.

        Args:
            in_data (dict): Raw event structure from packager

        Raises:
            BadDataError: If all packets are bad

        Returns:
            split_in_data (list): List of validated packets
        """
        split_in_data = self._split_rawdata_into_packets(in_data["rawdata"])
        for packet in split_in_data:
            if len(packet) != self._packet_size:
                split_in_data.remove(packet)
        if split_in_data == []:
            raise BadDataError("Input package has no valid packets")
        return split_in_data

    def parse_digital_data_old(self, in_data) -> dict:
        """Parse the raw data from the board.

        Since the data packets are constant length, we can extract the data
        in place with matrix operations, speeding up the parsing.

        Args:
            in_data (bytearray): Raw data from the board

        Returns:
            Parsed event as a dict.

        Raises:
            BadDataException if no data is found or if the data contains errors.
        """
        try:
            input_packets = self._validate_input_package(in_data)
        except:
            raise

        num_packets = len(input_packets)
        raw_data = np.frombuffer(np.array(input_packets).flatten(), dtype=">H")
        raw_data = np.reshape(
            np.array(raw_data, dtype="uint16"), (num_packets, self._packet_size // 2)
        )

        abs_wind_mask = self._abs_wind_mask
        evt_wind_mask = self._evt_wind_mask
        evt_wind_shift = self._evt_wind_shift
        chan_shift = self._chan_shift
        chan_mask = self._chan_mask
        timing_mask = self._timing_mask
        timing_shift = self._timing_shift

        curr_event = {
            "window_labels": [[] for _ in range(self.params["channels"])],
            "evt_window_labels": [[] for _ in range(self.params["channels"])],
            "data": [[] for _ in range(self.params["channels"])],
            "timing": [[] for _ in range(self.params["channels"])],
            "time": [[] for _ in range(self.params["channels"])],
        }

        current_time_dict = defaultdict(int)
        for packet in raw_data:
            # If first data bytes are zero, channel probably disabled.
            # Disabled channel check.
            if packet[4] == 0 and packet[5] == 0:
                continue

            channel = (packet[0] >> chan_shift) & chan_mask
            abs_window = packet[3] & abs_wind_mask
            evt_window = (packet[3] >> evt_wind_shift) & evt_wind_mask
            timing = (int(packet[1] & timing_mask) << timing_shift) | int(
                packet[2] & timing_mask
            )

            time_multiplier = current_time_dict[channel]
            current_time_dict[channel] = time_multiplier + 1
            time = np.arange(
                time_multiplier * self.samples, (time_multiplier + 1) * self.samples
            )

            curr_event["window_labels"][channel].append(abs_window)
            curr_event["evt_window_labels"][channel].append(evt_window)
            curr_event["timing"][channel].append(timing)
            curr_event["data"][channel].extend(packet[4:])
            curr_event["time"][channel].extend(time)
        curr_event["data"] = [np.array(x) for x in curr_event["data"]]
        return curr_event

    def _split_rawdata_into_packets(self, rawdata: bytearray) -> List[bytearray]:
        """Splits raw event data into packets.

        Args:
            rawdata (bytearray): the raw data to split

        Returns:
            A list of raw packets.
        """
        return rawdata.split(self._stop_word)

    def _add_xaxis_to_event(self, event):
        """Adds an x-axis to the data.

        Based on the amount of channels and samples it will add a timeaxis
        per channel to the event dict.

        During certain readout modes the window labels are not aligning,
        This function accounts for that and moves the time axis accordingly.

        It uses the window labels to determine the time axis by finding the lowest
        window number and offsetting the time axis by that amount.
        The 0-time will be the first samples in the lowest window.

        Args:
            event (dict): The event to add the x-axis to.

        Returns:
            numpy array with sample numbers for each channel.
        """
        times = [list() for _ in range(self.channels)]
        samples = self.samples
        channels = self.channels
        try:
            minwind = np.min([x[0] for x in event["window_labels"] if len(x) > 0])
        except:
            LOGGER.exception("No window labels found")
        for chan in range(channels):
            if len(event["window_labels"][chan]) == 0:
                continue
            winds = np.array(event["window_labels"][chan], dtype="int32")
            winds -= minwind

            # Get the first derivative of the wind list
            res = np.array(list(map(operator.sub, winds[1:], winds[:-1])))
            looped_indices = np.argwhere(res <= 0)
            # For every roll back to the 0th window, add 64 to the remaining labels
            # Add 1 to index as derivative occurs one step before roll back
            for i in looped_indices:
                winds[i[0] + 1 :] += self.windows
            tim = (samples * np.repeat(winds, samples)) + (
                np.ones((len(winds), samples), dtype=int)
                * np.arange(0, samples, dtype=int)
            ).flatten()
            times[chan] = tim

        return [np.array(x) for x in times]

    def parse(self, raw_data) -> dict:
        """Parses raw_data into a dict.

        Parses the data from the raw event and strips the timestamp/id.
        An dict is created from the parsed data with the id preserved.

        Args:
            raw_data (bytes): Raw data packaged with a header and a stopword.
        Returns:
            dict.
        """
        event = {}
        try:
            event = self.parse_digital_data(raw_data)
        except (TypeError, Exception) as e_msg:
            LOGGER.exception("parse_digital_data failed: %s", e_msg)
            event = raw_data
        else:
            # Add the X-Axis.
            # event = self.datafiller(event)
            event["time"] = self._add_xaxis_to_event(event)
            event["created_at"] = raw_data.get("created_at", 0)
            event["pkg_num"] = raw_data.get("pkg_num", 0)
            event["event_num"] = raw_data.get("pkg_num", 0)

        event["name"] = None

        return event
